#
# fourcc are codes to specify the encoding method a audio or video string
# in RIFF file (.avi and .wav).
#
# The following lists come from mmpython project:
#    file: mmpython/video/fourcc.py
#    url:  http://sourceforge.net/projects/mmpython/
#

# List of codecs with no compression (compression rate=1.0)
UNCOMPRESSED_AUDIO = set((1, 3, 6))

audio_codec_name = {
    0x0000: 'Microsoft Unknown Wave Format',
    0x0001: 'Microsoft Pulse Code Modulation (PCM)',
    0x0002: 'Microsoft ADPCM',
    0x0003: 'IEEE Float',
    0x0004: 'Compaq Computer VSELP',
    0x0005: 'IBM CVSD',
    0x0006: 'Microsoft A-Law',
    0x0007: 'Microsoft mu-Law',
    0x0010: 'OKI ADPCM',
    0x0011: 'Intel DVI/IMA ADPCM',
    0x0012: 'Videologic MediaSpace ADPCM',
    0x0013: 'Sierra Semiconductor ADPCM',
    0x0014: 'Antex Electronics G.723 ADPCM',
    0x0015: 'DSP Solutions DigiSTD',
    0x0016: 'DSP Solutions DigiFIX',
    0x0017: 'Dialogic OKI ADPCM',
    0x0018: 'MediaVision ADPCM',
    0x0019: 'Hewlett-Packard CU',
    0x0020: 'Yamaha ADPCM',
    0x0021: 'Speech Compression Sonarc',
    0x0022: 'DSP Group TrueSpeech',
    0x0023: 'Echo Speech EchoSC1',
    0x0024: 'Audiofile AF36',
    0x0025: 'Audio Processing Technology APTX',
    0x0026: 'AudioFile AF10',
    0x0027: 'Prosody 1612',
    0x0028: 'LRC',
    0x0030: 'Dolby AC2',
    0x0031: 'Microsoft GSM 6.10',
    0x0032: 'MSNAudio',
    0x0033: 'Antex Electronics ADPCME',
    0x0034: 'Control Resources VQLPC',
    0x0035: 'DSP Solutions DigiREAL',
    0x0036: 'DSP Solutions DigiADPCM',
    0x0037: 'Control Resources CR10',
    0x0038: 'Natural MicroSystems VBXADPCM',
    0x0039: 'Crystal Semiconductor IMA ADPCM',
    0x003A: 'EchoSC3',
    0x003B: 'Rockwell ADPCM',
    0x003C: 'Rockwell Digit LK',
    0x003D: 'Xebec',
    0x0040: 'Antex Electronics G.721 ADPCM',
    0x0041: 'G.728 CELP',
    0x0042: 'MSG723',
    0x0050: 'Microsoft MPEG',
    0x0052: 'RT24',
    0x0053: 'PAC',
    0x0055: 'MPEG Layer 3',
    0x0059: 'Lucent G.723',
    0x0060: 'Cirrus',
    0x0061: 'ESPCM',
    0x0062: 'Voxware',
    0x0063: 'Canopus Atrac',
    0x0064: 'G.726 ADPCM',
    0x0065: 'G.722 ADPCM',
    0x0066: 'DSAT',
    0x0067: 'DSAT Display',
    0x0069: 'Voxware Byte Aligned',
    0x0070: 'Voxware AC8',
    0x0071: 'Voxware AC10',
    0x0072: 'Voxware AC16',
    0x0073: 'Voxware AC20',
    0x0074: 'Voxware MetaVoice',
    0x0075: 'Voxware MetaSound',
    0x0076: 'Voxware RT29HW',
    0x0077: 'Voxware VR12',
    0x0078: 'Voxware VR18',
    0x0079: 'Voxware TQ40',
    0x0080: 'Softsound',
    0x0081: 'Voxware TQ60',
    0x0082: 'MSRT24',
    0x0083: 'G.729A',
    0x0084: 'MVI MV12',
    0x0085: 'DF G.726',
    0x0086: 'DF GSM610',
    0x0088: 'ISIAudio',
    0x0089: 'Onlive',
    0x0091: 'SBC24',
    0x0092: 'Dolby AC3 SPDIF',
    0x0097: 'ZyXEL ADPCM',
    0x0098: 'Philips LPCBB',
    0x0099: 'Packed',
    0x0100: 'Rhetorex ADPCM',
    0x0101: 'IBM mu-law',
    0x0102: 'IBM A-law',
    0x0103: 'IBM AVC Adaptive Differential Pulse Code Modulation (ADPCM)',
    0x0111: 'Vivo G.723',
    0x0112: 'Vivo Siren',
    0x0123: 'Digital G.723',
    0x0140: 'Windows Media Video V8',
    0x0161: 'Windows Media Audio V7 / V8 / V9',
    0x0162: 'Windows Media Audio Professional V9',
    0x0163: 'Windows Media Audio Lossless V9',
    0x0200: 'Creative Labs ADPCM',
    0x0202: 'Creative Labs Fastspeech8',
    0x0203: 'Creative Labs Fastspeech10',
    0x0220: 'Quarterdeck',
    0x0300: 'Fujitsu FM Towns Snd',
    0x0400: 'BTV Digital',
    0x0680: 'VME VMPCM',
    0x1000: 'Olivetti GSM',
    0x1001: 'Olivetti ADPCM',
    0x1002: 'Olivetti CELP',
    0x1003: 'Olivetti SBC',
    0x1004: 'Olivetti OPR',
    0x1100: 'Lernout & Hauspie LH Codec',
    0x1400: 'Norris',
    0x1401: 'AT&T ISIAudio',
    0x1500: 'Soundspace Music Compression',
    0x2000: 'AC3',
    0x7A21: 'GSM-AMR (CBR, no SID)',
    0x7A22: 'GSM-AMR (VBR, including SID)',
    0xFFFF: 'Development codec'
}

video_fourcc_name = {
    '3IV1': '3ivx v1',
    '3IV2': '3ivx v2',
    'AASC': 'Autodesk Animator',
    'ABYR': 'Kensington ?ABYR?',
    'AEMI': 'Array VideoONE MPEG1-I Capture',
    'AFLC': 'Autodesk Animator FLC',
    'AFLI': 'Autodesk Animator FLI',
    'AMPG': 'Array VideoONE MPEG',
    'ANIM': 'Intel RDX (ANIM)',
    'AP41': 'AngelPotion Definitive',
    'ASV1': 'Asus Video v1',
    'ASV2': 'Asus Video v2',
    'ASVX': 'Asus Video 2.0 (audio)',
    'AUR2': 'Aura 2 Codec - YUV 4:2:2',
    'AURA': 'Aura 1 Codec - YUV 4:1:1',
    'BINK': 'RAD Game Tools Bink Video',
    'BT20': 'Conexant Prosumer Video',
    'BTCV': 'Conexant Composite Video Codec',
    'BW10': 'Data Translation Broadway MPEG Capture',
    'CC12': 'Intel YUV12',
    'CDVC': 'Canopus DV',
    'CFCC': 'Digital Processing Systems DPS Perception',
    'CGDI': 'Microsoft Office 97 Camcorder Video',
    'CHAM': 'Winnov Caviara Champagne',
    'CJPG': 'Creative WebCam JPEG',
    'CLJR': 'Cirrus Logic YUV 4 pixels',
    'CMYK': 'Common Data Format in Printing',
    'CPLA': 'Weitek 4:2:0 YUV Planar',
    'CRAM': 'Microsoft Video 1 (CRAM)',
    'CVID': 'Radius Cinepak',
    'CWLT': 'Microsoft Color WLT DIB',
    'CYUV': 'Creative Labs YUV',
    'CYUY': 'ATI YUV',
    'D261': 'H.261',
    'D263': 'H.263',
    'DIV3': 'DivX v3 MPEG-4 Low-Motion',
    'DIV4': 'DivX v3 MPEG-4 Fast-Motion',
    'DIV5': '?DIV5?',
    'DIVX': 'DivX v4',
    'divx': 'DivX',
    'DMB1': 'Matrox Rainbow Runner hardware MJPEG',
    'DMB2': 'Paradigm MJPEG',
    'DSVD': '?DSVD?',
    'DUCK': 'Duck True Motion 1.0',
    'DVAN': '?DVAN?',
    'DVE2': 'InSoft DVE-2 Videoconferencing',
    'dvsd': 'DV',
    'DVSD': 'DV',
    'DVX1': 'DVX1000SP Video Decoder',
    'DVX2': 'DVX2000S Video Decoder',
    'DVX3': 'DVX3000S Video Decoder',
    'DX50': 'DivX v5',
    'DXT1': 'Microsoft DirectX Compressed Texture (DXT1)',
    'DXT2': 'Microsoft DirectX Compressed Texture (DXT2)',
    'DXT3': 'Microsoft DirectX Compressed Texture (DXT3)',
    'DXT4': 'Microsoft DirectX Compressed Texture (DXT4)',
    'DXT5': 'Microsoft DirectX Compressed Texture (DXT5)',
    'DXTC': 'Microsoft DirectX Compressed Texture (DXTC)',
    'EKQ0': 'Elsa ?EKQ0?',
    'ELK0': 'Elsa ?ELK0?',
    'ESCP': 'Eidos Escape',
    'ETV1': 'eTreppid Video ETV1',
    'ETV2': 'eTreppid Video ETV2',
    'ETVC': 'eTreppid Video ETVC',
    'FLJP': 'D-Vision Field Encoded Motion JPEG',
    'FRWA': 'SoftLab-Nsk Forward Motion JPEG w/ alpha channel',
    'FRWD': 'SoftLab-Nsk Forward Motion JPEG',
    'FVF1': 'Iterated Systems Fractal Video Frame',
    'GLZW': 'Motion LZW (gabest@freemail.hu)',
    'GPEG': 'Motion JPEG (gabest@freemail.hu)',
    'GWLT': 'Microsoft Greyscale WLT DIB',
    'H260': 'Intel ITU H.260 Videoconferencing',
    'H261': 'Intel ITU H.261 Videoconferencing',
    'H262': 'Intel ITU H.262 Videoconferencing',
    'H263': 'Intel ITU H.263 Videoconferencing',
    'H264': 'Intel ITU H.264 Videoconferencing',
    'H265': 'Intel ITU H.265 Videoconferencing',
    'H266': 'Intel ITU H.266 Videoconferencing',
    'H267': 'Intel ITU H.267 Videoconferencing',
    'H268': 'Intel ITU H.268 Videoconferencing',
    'H269': 'Intel ITU H.269 Videoconferencing',
    'HFYU': 'Huffman Lossless Codec',
    'HMCR': 'Rendition Motion Compensation Format (HMCR)',
    'HMRR': 'Rendition Motion Compensation Format (HMRR)',
    'i263': 'Intel ITU H.263 Videoconferencing (i263)',
    'I420': 'Intel Indeo 4',
    'IAN ': 'Intel RDX',
    'ICLB': 'InSoft CellB Videoconferencing',
    'IGOR': 'Power DVD',
    'IJPG': 'Intergraph JPEG',
    'ILVC': 'Intel Layered Video',
    'ILVR': 'ITU-T H.263+',
    'IPDV': 'I-O Data Device Giga AVI DV Codec',
    'IR21': 'Intel Indeo 2.1',
    'IRAW': 'Intel YUV Uncompressed',
    'IV30': 'Ligos Indeo 3.0',
    'IV31': 'Ligos Indeo 3.1',
    'IV32': 'Ligos Indeo 3.2',
    'IV33': 'Ligos Indeo 3.3',
    'IV34': 'Ligos Indeo 3.4',
    'IV35': 'Ligos Indeo 3.5',
    'IV36': 'Ligos Indeo 3.6',
    'IV37': 'Ligos Indeo 3.7',
    'IV38': 'Ligos Indeo 3.8',
    'IV39': 'Ligos Indeo 3.9',
    'IV40': 'Ligos Indeo Interactive 4.0',
    'IV41': 'Ligos Indeo Interactive 4.1',
    'IV42': 'Ligos Indeo Interactive 4.2',
    'IV43': 'Ligos Indeo Interactive 4.3',
    'IV44': 'Ligos Indeo Interactive 4.4',
    'IV45': 'Ligos Indeo Interactive 4.5',
    'IV46': 'Ligos Indeo Interactive 4.6',
    'IV47': 'Ligos Indeo Interactive 4.7',
    'IV48': 'Ligos Indeo Interactive 4.8',
    'IV49': 'Ligos Indeo Interactive 4.9',
    'IV50': 'Ligos Indeo Interactive 5.0',
    'JBYR': 'Kensington ?JBYR?',
    'JPEG': 'Still Image JPEG DIB',
    'JPGL': 'Webcam JPEG Light?',
    'KMVC': 'Karl Morton\'s Video Codec',
    'LEAD': 'LEAD Video Codec',
    'Ljpg': 'LEAD MJPEG Codec',
    'M261': 'Microsoft H.261',
    'M263': 'Microsoft H.263',
    'M4S2': 'Microsoft MPEG-4 (M4S2)',
    'm4s2': 'Microsoft MPEG-4 (m4s2)',
    'MC12': 'ATI Motion Compensation Format (MC12)',
    'MCAM': 'ATI Motion Compensation Format (MCAM)',
    'MJ2C': 'Morgan Multimedia Motion JPEG2000',
    'mJPG': 'IBM Motion JPEG w/ Huffman Tables',
    'MJPG': 'Motion JPEG DIB',
    'MP42': 'Microsoft MPEG-4 (low-motion)',
    'MP43': 'Microsoft MPEG-4 (fast-motion)',
    'MP4S': 'Microsoft MPEG-4 (MP4S)',
    'mp4s': 'Microsoft MPEG-4 (mp4s)',
    'MPEG': 'MPEG 1 Video I-Frame',
    'MPG4': 'Microsoft MPEG-4 Video High Speed Compressor',
    'MPGI': 'Sigma Designs MPEG',
    'MRCA': 'FAST Multimedia Mrcodec',
    'MRLE': 'Microsoft Run Length Encoding (RLE)',
    'MSVC': 'Microsoft Video 1',
    'MTX1': 'Matrox ?MTX1?',
    'MTX2': 'Matrox ?MTX2?',
    'MTX3': 'Matrox ?MTX3?',
    'MTX4': 'Matrox ?MTX4?',
    'MTX5': 'Matrox ?MTX5?',
    'MTX6': 'Matrox ?MTX6?',
    'MTX7': 'Matrox ?MTX7?',
    'MTX8': 'Matrox ?MTX8?',
    'MTX9': 'Matrox ?MTX9?',
    'MV12': '?MV12?',
    'MWV1': 'Aware Motion Wavelets',
    'nAVI': '?nAVI?',
    'NTN1': 'Nogatech Video Compression 1',
    'NVS0': 'nVidia GeForce Texture (NVS0)',
    'NVS1': 'nVidia GeForce Texture (NVS1)',
    'NVS2': 'nVidia GeForce Texture (NVS2)',
    'NVS3': 'nVidia GeForce Texture (NVS3)',
    'NVS4': 'nVidia GeForce Texture (NVS4)',
    'NVS5': 'nVidia GeForce Texture (NVS5)',
    'NVT0': 'nVidia GeForce Texture (NVT0)',
    'NVT1': 'nVidia GeForce Texture (NVT1)',
    'NVT2': 'nVidia GeForce Texture (NVT2)',
    'NVT3': 'nVidia GeForce Texture (NVT3)',
    'NVT4': 'nVidia GeForce Texture (NVT4)',
    'NVT5': 'nVidia GeForce Texture (NVT5)',
    'PDVC': 'I-O Data Device Digital Video Capture DV codec',
    'PGVV': 'Radius Video Vision',
    'PHMO': 'IBM Photomotion',
    'PIM1': 'Pegasus Imaging ?PIM1?',
    'PIM2': 'Pegasus Imaging ?PIM2?',
    'PIMJ': 'Pegasus Imaging Lossless JPEG',
    'PVEZ': 'Horizons Technology PowerEZ',
    'PVMM': 'PacketVideo Corporation MPEG-4',
    'PVW2': 'Pegasus Imaging Wavelet Compression',
    'QPEG': 'Q-Team QPEG 1.0',
    'qpeq': 'Q-Team QPEG 1.1',
    'RGBT': 'Computer Concepts 32-bit support',
    'RLE ': 'Microsoft Run Length Encoder',
    'RLE4': 'Run Length Encoded 4',
    'RLE8': 'Run Length Encoded 8',
    'RT21': 'Intel Real Time Video 2.1',
    'rv20': 'RealVideo G2',
    'rv30': 'RealVideo 8',
    'RVX ': 'Intel RDX (RVX )',
    's422': 'Tekram VideoCap C210 YUV 4:2:2',
    'SDCC': 'Sun Communication Digital Camera Codec',
    'SFMC': 'CrystalNet Surface Fitting Method',
    'SMSC': 'Radius SMSC',
    'SMSD': 'Radius SMSD',
    'smsv': 'WorldConnect Wavelet Video',
    'SPIG': 'Radius Spigot',
    'SPLC': 'Splash Studios ACM Audio Codec',
    'SQZ2': 'Microsoft VXTreme Video Codec V2',
    'STVA': 'ST CMOS Imager Data (Bayer)',
    'STVB': 'ST CMOS Imager Data (Nudged Bayer)',
    'STVC': 'ST CMOS Imager Data (Bunched)',
    'STVX': 'ST CMOS Imager Data (Extended CODEC Data Format)',
    'STVY': 'ST CMOS Imager Data (Extended CODEC Data Format with Correction Data)',
    'SV10': 'Sorenson Video R1',
    'SVQ1': 'Sorenson Video R3',
    'TLMS': 'TeraLogic Motion Intraframe Codec (TLMS)',
    'TLST': 'TeraLogic Motion Intraframe Codec (TLST)',
    'TM20': 'Duck TrueMotion 2.0',
    'TM2X': 'Duck TrueMotion 2X',
    'TMIC': 'TeraLogic Motion Intraframe Codec (TMIC)',
    'TMOT': 'Horizons Technology TrueMotion S',
    'tmot': 'Horizons TrueMotion Video Compression',
    'TR20': 'Duck TrueMotion RealTime 2.0',
    'TSCC': 'TechSmith Screen Capture Codec',
    'TV10': 'Tecomac Low-Bit Rate Codec',
    'TY0N': 'Trident ?TY0N?',
    'TY2C': 'Trident ?TY2C?',
    'TY2N': 'Trident ?TY2N?',
    'UCOD': 'eMajix.com ClearVideo',
    'ULTI': 'IBM Ultimotion',
    'UYVY': 'UYVY 4:2:2 byte ordering',
    'V261': 'Lucent VX2000S',
    'V422': '24 bit YUV 4:2:2 Format',
    'V655': '16 bit YUV 4:2:2 Format',
    'VCR1': 'ATI VCR 1.0',
    'VCR2': 'ATI VCR 2.0',
    'VCR3': 'ATI VCR 3.0',
    'VCR4': 'ATI VCR 4.0',
    'VCR5': 'ATI VCR 5.0',
    'VCR6': 'ATI VCR 6.0',
    'VCR7': 'ATI VCR 7.0',
    'VCR8': 'ATI VCR 8.0',
    'VCR9': 'ATI VCR 9.0',
    'VDCT': 'Video Maker Pro DIB',
    'VDOM': 'VDOnet VDOWave',
    'VDOW': 'VDOnet VDOLive (H.263)',
    'VDTZ': 'Darim Vison VideoTizer YUV',
    'VGPX': 'VGPixel Codec',
    'VIDS': 'Vitec Multimedia YUV 4:2:2 CCIR 601 for V422',
    'VIFP': '?VIFP?',
    'VIVO': 'Vivo H.263 v2.00',
    'vivo': 'Vivo H.263',
    'VIXL': 'Miro Video XL',
    'VLV1': 'Videologic VLCAP.DRV',
    'VP30': 'On2 VP3.0',
    'VP31': 'On2 VP3.1',
    'VX1K': 'VX1000S Video Codec',
    'VX2K': 'VX2000S Video Codec',
    'VXSP': 'VX1000SP Video Codec',
    'WBVC': 'Winbond W9960',
    'WHAM': 'Microsoft Video 1 (WHAM)',
    'WINX': 'Winnov Software Compression',
    'WJPG': 'AverMedia Winbond JPEG',
    'WMV1': 'Windows Media Video V7',
    'WMV2': 'Windows Media Video V8',
    'WMV3': 'Windows Media Video V9',
    'WNV1': 'Winnov Hardware Compression',
    'x263': 'Xirlink H.263',
    'XLV0': 'NetXL Video Decoder',
    'XMPG': 'Xing MPEG (I-Frame only)',
    'XVID': 'XviD MPEG-4',
    'XXAN': '?XXAN?',
    'Y211': 'YUV 2:1:1 Packed',
    'Y411': 'YUV 4:1:1 Packed',
    'Y41B': 'YUV 4:1:1 Planar',
    'Y41P': 'PC1 4:1:1',
    'Y41T': 'PC1 4:1:1 with transparency',
    'Y42B': 'YUV 4:2:2 Planar',
    'Y42T': 'PCI 4:2:2 with transparency',
    'Y8  ': 'Grayscale video',
    'YC12': 'Intel YUV12 Codec',
    'YUV8': 'Winnov Caviar YUV8',
    'YUV9': 'Intel YUV9',
    'YUY2': 'Uncompressed YUV 4:2:2',
    'YUYV': 'Canopus YUV',
    'YV12': 'YVU12 Planar',
    'YVU9': 'Intel YVU9 Planar',
    'YVYU': 'YVYU 4:2:2 byte ordering',
    'ZLIB': '?ZLIB?',
    'ZPEG': 'Metheus Video Zipper'
}



"""
New Technology File System (NTFS) file system parser.

Sources:
- The NTFS documentation
  http://www.linux-ntfs.org/
- NTFS-3G driver
  http://www.ntfs-3g.org/

Creation date: 3rd january 2007
Author: Victor Stinner
"""

from hachoir.parser import Parser
from hachoir.field import (FieldSet, Enum,
                           UInt8, UInt16, UInt32, UInt64, TimestampWin64,
                           String, Bytes, Bit, Bits,
                           NullBits, NullBytes, PaddingBytes, RawBytes)
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.core.text_handler import textHandler, hexadecimal, filesizeHandler
from hachoir.core.tools import humanFilesize, createDict
from hachoir.parser.common.msdos import MSDOSFileAttr32

SECTOR_SIZE = 512


class BiosParameterBlock(FieldSet):
    """
    BIOS parameter block (bpb) structure
    """
    static_size = 25 * 8
    MEDIA_TYPE = {0xf8: "Hard disk"}

    def createFields(self):
        yield UInt16(self, "bytes_per_sector", "Size of a sector in bytes")
        yield UInt8(self, "sectors_per_cluster", "Size of a cluster in sectors")
        yield NullBytes(self, "reserved_sectors", 2)
        yield NullBytes(self, "fats", 1)
        yield NullBytes(self, "root_entries", 2)
        yield NullBytes(self, "sectors", 2)
        yield Enum(UInt8(self, "media_type"), self.MEDIA_TYPE)
        yield NullBytes(self, "sectors_per_fat", 2)
        yield UInt16(self, "sectors_per_track")
        yield UInt16(self, "heads")
        yield UInt32(self, "hidden_sectors")
        yield NullBytes(self, "large_sectors", 4)

    def validate(self):
        if self["bytes_per_sector"].value not in (256, 512, 1024, 2048, 4096):
            return "Invalid sector size (%u bytes)" % \
                self["bytes_per_sector"].value
        if self["sectors_per_cluster"].value not in (1, 2, 4, 8, 16, 32, 64, 128):
            return "Invalid cluster size (%u sectors)" % \
                self["sectors_per_cluster"].value
        return ""


class MasterBootRecord(FieldSet):
    static_size = 512 * 8

    def createFields(self):
        yield Bytes(self, "jump", 3, "Intel x86 jump instruction")
        yield String(self, "name", 8)
        yield BiosParameterBlock(self, "bios", "BIOS parameters")

        yield textHandler(UInt8(self, "physical_drive", "(0x80)"), hexadecimal)
        yield NullBytes(self, "current_head", 1)
        yield textHandler(UInt8(self, "ext_boot_sig", "Extended boot signature (0x80)"), hexadecimal)
        yield NullBytes(self, "unused", 1)

        yield UInt64(self, "nb_sectors")
        yield UInt64(self, "mft_cluster", "Cluster location of MFT data")
        yield UInt64(self, "mftmirr_cluster", "Cluster location of copy of MFT")
        yield UInt8(self, "cluster_per_mft", "MFT record size in clusters")
        yield NullBytes(self, "reserved[]", 3)
        yield UInt8(self, "cluster_per_index", "Index block size in clusters")
        yield NullBytes(self, "reserved[]", 3)
        yield textHandler(UInt64(self, "serial_number"), hexadecimal)
        yield textHandler(UInt32(self, "checksum", "Boot sector checksum"), hexadecimal)
        yield Bytes(self, "boot_code", 426)
        yield Bytes(self, "mbr_magic", 2, r"Master boot record magic number (\x55\xAA)")

    def createDescription(self):
        size = self["nb_sectors"].value * self["bios/bytes_per_sector"].value
        return "NTFS Master Boot Record (%s)" % humanFilesize(size)


class MFT_Flags(FieldSet):
    static_size = 16

    def createFields(self):
        yield Bit(self, "in_use")
        yield Bit(self, "is_directory")
        yield NullBits(self, "padding", 14)


class RunListEntry(FieldSet):
    def createFields(self):
        header = UInt8(self, "header")
        yield header
        if header.value == 0:
            return

        offlen = header.value >> 4
        lenlen = header.value & 0xf

        if lenlen:
            yield Bits(self, "length", lenlen * 8)
        if offlen:
            yield Bits(self, "offset", offlen * 8)

    def createValue(self):
        if 'length' in self:
            length = self['length'].value
        else:
            length = 0

        if 'offset' in self:
            offset = self['offset'].value
            offlen = self['offset'].size
            if offset & (1 << (offlen - 1)):
                offset -= 1 << offlen
        else:
            offset = 0

        return (length, offset)


class RunList(FieldSet):
    def createFields(self):
        while 1:
            entry = RunListEntry(self, "entry[]")
            yield entry
            if entry['header'].value == 0:
                break

    def createValue(self):
        runs = []
        curoff = 0
        for entry in self.array('entry'):
            length, offset = entry.value
            curoff += offset
            if entry['header'].value != 0:
                runs.append((length, curoff))
        return runs


class Attribute(FieldSet):
    # --- Common code ---

    def __init__(self, *args):
        FieldSet.__init__(self, *args)
        self._size = self["size"].value * 8
        type = self["type"].value
        if type in self.ATTR_INFO:
            self._name = self.ATTR_INFO[type][0]
            self._parser = self.ATTR_INFO[type][2]

    def createFields(self):
        yield Enum(textHandler(UInt32(self, "type"), hexadecimal), self.ATTR_NAME)
        yield UInt32(self, "size")
        yield UInt8(self, "non_resident", "Non-resident flag")
        yield UInt8(self, "name_length", "Name length in bytes")
        yield UInt16(self, "name_offset", "Name offset")
        yield UInt16(self, "flags")
        yield textHandler(UInt16(self, "attribute_id"), hexadecimal)
        if self['non_resident'].value:
            yield UInt64(self, "runlist_start", "Starting Virtual Cluster Number of the runlist")
            yield UInt64(self, "runlist_stop", "Ending Virtual Cluster Number of the runlist")
            yield UInt16(self, "runlist_offset", "Offset to the runlist")
            yield UInt16(self, "compression_unit", "Compression unit size")
            yield UInt32(self, "unused[]")
            yield UInt64(self, "size_allocated", "Allocated size of the attribute content")
            yield UInt64(self, "size_actual", "Actual size of the attribute content")
            yield UInt64(self, "size_initialized", "Initialized size of the attribute content")

            if self['name_length'].value:
                padding = self.seekByte(self['name_offset'].value)
                if padding:
                    yield padding
                yield String(self, "name", self['name_length'].value * 2, charset="UTF-16-LE")

            padding = self.seekByte(self['runlist_offset'].value)
            if padding:
                yield padding
            yield RunList(self, "runlist")

        else:
            yield UInt32(self, "length_attr", "Length of the Attribute")
            yield UInt16(self, "offset_attr", "Offset of the Attribute")
            yield UInt8(self, "indexed_flag")

            if self['name_length'].value:
                padding = self.seekByte(self['name_offset'].value)
                if padding:
                    yield padding
                yield String(self, "name", self['name_length'].value * 2, charset="UTF-16-LE")

            padding = self.seekByte(self['offset_attr'].value)
            if padding:
                yield padding

            if self._parser:
                yield from self._parser(self)
            else:
                size = self["length_attr"].value
                if size:
                    yield RawBytes(self, "data", size)

        size = (self.size - self.current_size) // 8
        if size:
            yield PaddingBytes(self, "end_padding", size)

    def createDescription(self):
        return "Attribute %s" % self["type"].display
    FILENAME_NAMESPACE = {
        0: "POSIX",
        1: "Win32",
        2: "DOS",
        3: "Win32 & DOS",
    }

    # --- Parser specific to a type ---
    def parseStandardInfo(self):
        yield TimestampWin64(self, "ctime", "File Creation")
        yield TimestampWin64(self, "atime", "File Altered")
        yield TimestampWin64(self, "mtime", "MFT Changed")
        yield TimestampWin64(self, "rtime", "File Read")
        yield MSDOSFileAttr32(self, "file_attr", "DOS File Permissions")
        yield UInt32(self, "max_version", "Maximum Number of Versions")
        yield UInt32(self, "version", "Version Number")
        yield UInt32(self, "class_id")
        yield UInt32(self, "owner_id")
        yield UInt32(self, "security_id")
        yield filesizeHandler(UInt64(self, "quota_charged", "Quota Charged"))
        yield UInt64(self, "usn", "Update Sequence Number (USN)")

    def parseFilename(self):
        yield UInt64(self, "ref", "File reference to the parent directory")
        yield TimestampWin64(self, "ctime", "File Creation")
        yield TimestampWin64(self, "atime", "File Altered")
        yield TimestampWin64(self, "mtime", "MFT Changed")
        yield TimestampWin64(self, "rtime", "File Read")
        yield filesizeHandler(UInt64(self, "alloc_size", "Allocated size of the file"))
        yield filesizeHandler(UInt64(self, "real_size", "Real size of the file"))
        yield UInt32(self, "file_flags")
        yield UInt32(self, "file_flags2", "Used by EAs and Reparse")
        yield UInt8(self, "filename_length", "Filename length in characters")
        yield Enum(UInt8(self, "filename_namespace"), self.FILENAME_NAMESPACE)
        size = self["filename_length"].value * 2
        if size:
            yield String(self, "filename", size, charset="UTF-16-LE")

    def parseData(self):
        size = (self.size - self.current_size) // 8
        if size:
            yield Bytes(self, "data", size)

    def parseBitmap(self):
        size = (self.size - self.current_size)
        for index in range(size):
            yield Bit(self, "bit[]")

    # --- Type information ---
    ATTR_INFO = {
        0x10: ('standard_info', 'STANDARD_INFORMATION ', parseStandardInfo),
        0x20: ('attr_list', 'ATTRIBUTE_LIST ', None),
        0x30: ('filename[]', 'FILE_NAME ', parseFilename),
        0x40: ('vol_ver', 'VOLUME_VERSION', None),
        0x50: ('security', 'SECURITY_DESCRIPTOR ', None),
        0x60: ('vol_name', 'VOLUME_NAME ', None),
        0x70: ('vol_info', 'VOLUME_INFORMATION ', None),
        0x80: ('data', 'DATA ', parseData),
        0x90: ('index_root', 'INDEX_ROOT ', None),
        0xA0: ('index_alloc', 'INDEX_ALLOCATION ', None),
        0xB0: ('bitmap', 'BITMAP ', parseBitmap),
        0xC0: ('reparse', 'REPARSE_POINT ', None),
        0xD0: ('ea_info', 'EA_INFORMATION ', None),
        0xE0: ('ea', 'EA ', None),
        0xF0: ('prop_set', 'PROPERTY_SET', None),
        0x100: ('log_util', 'LOGGED_UTILITY_STREAM', None),
    }
    ATTR_NAME = createDict(ATTR_INFO, 1)


class File(FieldSet):
    #    static_size = 48*8

    def __init__(self, *args):
        FieldSet.__init__(self, *args)
        self._size = self["bytes_allocated"].value * 8

    def createFields(self):
        yield Bytes(self, "signature", 4, "Usually the magic is 'FILE'")
        yield UInt16(self, "usa_offset", "Update Sequence Array offset")
        yield UInt16(self, "usa_count", "Update Sequence Array count")
        yield UInt64(self, "lsn", "$LogFile sequence number for this record")
        yield UInt16(self, "sequence_number", "Number of times this mft record has been reused")
        yield UInt16(self, "link_count", "Number of hard links")
        yield UInt16(self, "attrs_offset", "Byte offset to the first attribute")
        yield MFT_Flags(self, "flags")
        yield UInt32(self, "bytes_in_use", "Number of bytes used in this record")
        yield UInt32(self, "bytes_allocated", "Number of bytes allocated for this record")
        yield UInt64(self, "base_mft_record")
        yield UInt16(self, "next_attr_instance")

        # The below fields are specific to NTFS 3.1+ (Windows XP and above)
        yield NullBytes(self, "reserved", 2)
        yield UInt32(self, "mft_record_number", "Number of this mft record")

        if self["usa_offset"].value:
            padding = self.seekByte(self["usa_offset"].value, relative=True)
            if padding:
                yield padding

            yield UInt16(self, "usa_number")
            for i in range(self["usa_count"].value):
                yield UInt16(self, "usa_value[]")

        padding = self.seekByte(self["attrs_offset"].value, relative=True)
        if padding:
            yield padding

        while not self.eof:
            addr = self.absolute_address + self.current_size
            if self.stream.readBytes(addr, 4) == b"\xFF\xFF\xFF\xFF":
                yield Bytes(self, "attr_end_marker", 8)
                break
            yield Attribute(self, "attr[]")

        size = self["bytes_in_use"].value - self.current_size // 8
        if size:
            yield RawBytes(self, "end_rawdata", size)

        size = (self.size - self.current_size) // 8
        if size:
            yield RawBytes(self, "end_padding", size, "Unused but allocated bytes")

    def createDescription(self):
        text = "File"
        fileattr = None
        for fn in self.array('filename'):
            if 'filename' not in fn and fileattr is not None:
                # Prefer file attributes with filenames
                continue
            if 'Win32' in fn['filename_namespace'].display:
                # Prefer Win32 filename
                fileattr = fn
                break
            else:
                fileattr = fn

        if fileattr:
            if 'filename' in fileattr:
                text += ' "%s"' % fileattr["filename"].value
            text += ' (%s)' % fileattr["real_size"].display
        if "standard_info/file_attr" in self:
            text += ', %s' % self["standard_info/file_attr"].display
        return text


class NTFS(Parser):
    MAGIC = b"\xEB\x52\x90NTFS    "
    PARSER_TAGS = {
        "id": "ntfs",
        "category": "file_system",
        "description": "NTFS file system",
        "min_size": 1024 * 8,
        "magic": ((MAGIC, 0),),
    }
    endian = LITTLE_ENDIAN
    _cluster_size = None

    def validate(self):
        if self.stream.readBytes(0, len(self.MAGIC)) != self.MAGIC:
            return "Invalid magic string"
        err = self["mbr/bios"].validate()
        if err:
            return err
        return True

    def createFields(self):
        yield MasterBootRecord(self, "mbr")

        bios = self["mbr/bios"]
        cluster_size = bios["sectors_per_cluster"].value * \
            bios["bytes_per_sector"].value
        offset = self["mbr/mft_cluster"].value * cluster_size
        padding = self.seekByte(offset, relative=False)
        if padding:
            yield padding
        for index in range(1000):
            yield File(self, "file[]")

        size = (self.size - self.current_size) // 8
        if size:
            yield RawBytes(self, "end", size)



"""
TIFF image parser.

Authors: Victor Stinner, Sebastien Ponce, Robert Xiao
Creation date: 30 september 2006
"""

from hachoir.parser import Parser
from hachoir.field import SeekableFieldSet, RootSeekableFieldSet, Bytes
from hachoir.core.endian import LITTLE_ENDIAN, BIG_ENDIAN
from hachoir.parser.image.exif import TIFF, IFD


def getStrips(ifd):
    data = {}
    for i, entry in enumerate(ifd.array('entry')):
        data[entry['tag'].display] = entry
    # image data
    if "StripOffsets" in data and "StripByteCounts" in data:
        offs = ifd.getEntryValues(data["StripOffsets"])
        bytes = ifd.getEntryValues(data["StripByteCounts"])
        for off, byte in zip(offs, bytes):
            yield off.value, byte.value

    # image data
    if "TileOffsets" in data and "TileByteCounts" in data:
        offs = ifd.getEntryValues(data["TileOffsets"])
        bytes = ifd.getEntryValues(data["TileByteCounts"])
        for off, byte in zip(offs, bytes):
            yield off.value, byte.value


class ImageFile(SeekableFieldSet):

    def __init__(self, parent, name, description, ifd):
        SeekableFieldSet.__init__(self, parent, name, description, None)
        self._ifd = ifd

    def createFields(self):
        for off, byte in getStrips(self._ifd):
            self.seekByte(off, relative=False)
            field = Bytes(self, "strip[]", byte)
            yield field


class TiffFile(RootSeekableFieldSet, Parser):
    PARSER_TAGS = {
        "id": "tiff",
        "category": "image",
        "file_ext": ("tif", "tiff"),
        "mime": ("image/tiff",),
        "min_size": 8 * 8,
        "magic": ((b"II\x2A\0", 0), (b"MM\0\x2A", 0)),
        "description": "TIFF picture"
    }

    # Correct endian is set in constructor
    endian = LITTLE_ENDIAN

    def __init__(self, stream, **args):
        RootSeekableFieldSet.__init__(
            self, None, "root", stream, None, stream.askSize(self))
        if self.stream.readBytes(0, 2) == b"MM":
            self.endian = BIG_ENDIAN
        Parser.__init__(self, stream, **args)

    def validate(self):
        endian = self.stream.readBytes(0, 2)
        if endian not in (b"MM", b"II"):
            return "Invalid endian (%r)" % endian
        if self["version"].value != 42:
            return "Unknown TIFF version"
        return True

    def createFields(self):
        yield from TIFF(self)

        for ifd in self:
            if not isinstance(ifd, IFD):
                continue
            offs = (off for off, byte in getStrips(ifd))
            self.seekByte(min(offs), relative=False)
            image = ImageFile(self, "image[]", "Image File", ifd)
            yield image



import functools
from hachoir.core.iso639 import ISO639_2


@functools.total_ordering
class Language:

    def __init__(self, code):
        code = str(code)
        if code not in ISO639_2:
            raise ValueError("Invalid language code: %r" % code)
        self.code = code

    def __eq__(self, other):
        if other.__class__ != Language:
            return NotImplemented
        return self.code == other.code

    def __lt__(self, other):
        if other.__class__ != Language:
            return NotImplemented
        return self.code < other.code

    def __str__(self):
        return ISO639_2[self.code]

    def __repr__(self):
        return "<Language '%s', code=%r>" % (str(self), self.code)



"""
Abstract Syntax Notation One (ASN.1) parser.

Technical informations:
* PER standard
  http://www.tu.int/ITU-T/studygroups/com17/languages/X.691-0207.pdf
* Python library
  http://pyasn1.sourceforge.net/
* Specification of Abstract Syntax Notation One (ASN.1)
  ISO/IEC 8824:1990 Information Technology
* Specification of Basic Encoding Rules (BER) for ASN.1
  ISO/IEC 8825:1990 Information Technology
* OpenSSL asn1parser, use command:
  openssl asn1parse -i -inform DER -in file.der
* ITU-U recommendations:
  http://www.itu.int/rec/T-REC-X/en
  (X.680, X.681, X.682, X.683, X.690, X.691, X.692, X.693, X.694)
* dumpasn1
  http://www.cs.auckland.ac.nz/~pgut001/dumpasn1.c

General information:
* Wikipedia (english) article
  http://en.wikipedia.org/wiki/Abstract_Syntax_Notation_One
* ASN.1 information site
  http://asn1.elibel.tm.fr/en/
* ASN.1 consortium
  http://www.asn1.org/

Encodings:
* Basic Encoding Rules (BER)
* Canonical Encoding Rules (CER) -- DER derivative that is not widely used
* Distinguished Encoding Rules (DER) -- used for encrypted applications
* XML Encoding Rules (XER)
* Packed Encoding Rules (PER) -- result in the fewest number of bytes
* Generic String Encoding Rules (GSER)
=> Are encodings compatibles? Which encodings are supported??

Author: Victor Stinner
Creation date: 24 september 2006
"""

from hachoir.parser import Parser
from hachoir.field import (FieldSet,
                           FieldError, ParserError,
                           Bit, Bits, Bytes, UInt8, GenericInteger, String,
                           Field, Enum, RawBytes)
from hachoir.core.endian import BIG_ENDIAN
from hachoir.core.tools import createDict, humanDatetime
from hachoir.stream import InputStreamError
from hachoir.core.text_handler import textHandler

# --- Field parser ---


class ASNInteger(Field):
    """
    Integer: two cases:
    - first byte in 0..127: it's the value
    - first byte in 128..255: byte & 127 is the number of bytes,
      next bytes are the value
    """

    def __init__(self, parent, name, description=None):
        Field.__init__(self, parent, name, 8, description)
        stream = self._parent.stream
        addr = self.absolute_address
        value = stream.readBits(addr, 8, BIG_ENDIAN)
        if 128 <= value:
            nbits = (value & 127) * 8
            if not nbits:
                raise ParserError("ASN.1: invalid ASN integer size (zero)")
            if 64 < nbits:
                # Arbitrary limit to catch errors
                raise ParserError("ASN.1: ASN integer is limited to 64 bits")
            self._size = 8 + nbits
            value = stream.readBits(addr + 8, nbits, BIG_ENDIAN)
        self.createValue = lambda: value


class OID_Integer(Bits):

    def __init__(self, parent, name, description=None):
        Bits.__init__(self, parent, name, 8, description)
        stream = self._parent.stream
        addr = self.absolute_address
        size = 8
        value = 0
        byte = stream.readBits(addr, 8, BIG_ENDIAN)
        value = byte & 127
        while 128 <= byte:
            addr += 8
            size += 8
            if 64 < size:
                # Arbitrary limit to catch errors
                raise ParserError(
                    "ASN.1: Object identifier is limited 64 bits")
            byte = stream.readBits(addr, 8, BIG_ENDIAN)
            value = (value << 7) + (byte & 127)
        self._size = size
        self.createValue = lambda: value


def readSequence(self, content_size):
    while self.current_size < self.size:
        yield Object(self, "item[]")


def readSet(self, content_size):
    yield Object(self, "value", size=content_size * 8)


def readASCIIString(self, content_size):
    yield String(self, "value", content_size, charset="ASCII")


def readUTF8String(self, content_size):
    yield String(self, "value", content_size, charset="UTF-8")


def readBMPString(self, content_size):
    yield String(self, "value", content_size, charset="UTF-16")


def readBitString(self, content_size):
    yield UInt8(self, "padding_size", description="Number of unused bits")
    if content_size > 1:
        yield Bytes(self, "value", content_size - 1)


def readOctetString(self, content_size):
    yield Bytes(self, "value", content_size)


def formatObjectID(fieldset):
    text = [fieldset["first"].display]
    items = [field for field in fieldset if field.name.startswith("item[")]
    text.extend(str(field.value) for field in items)
    return ".".join(text)


def readObjectID(self, content_size):
    yield textHandler(UInt8(self, "first"), formatFirstObjectID)
    while self.current_size < self.size:
        yield OID_Integer(self, "item[]")


def readBoolean(self, content_size):
    if content_size != 1:
        raise ParserError(
            "Overlong boolean: got %s bytes, expected 1 byte" % content_size)
    yield textHandler(UInt8(self, "value"), lambda field: str(bool(field.value)))


def readInteger(self, content_size):
    # Always signed?
    yield GenericInteger(self, "value", True, content_size * 8)

# --- Format ---


def formatFirstObjectID(field):
    value = field.value
    return "%u.%u" % (value // 40, value % 40)


def formatValue(fieldset):
    return fieldset["value"].display


def formatUTCTime(fieldset):
    import datetime
    value = fieldset["value"].value
    year = int(value[0:2])
    if year < 50:
        year += 2000
    else:
        year += 1900
    month = int(value[2:4])
    day = int(value[4:6])
    hour = int(value[6:8])
    minute = int(value[8:10])
    if value[-1] == "Z":
        second = int(value[10:12])
        dt = datetime.datetime(year, month, day, hour, minute, second)
    else:
        # Skip timezone...
        dt = datetime.datetime(year, month, day, hour, minute)
    return humanDatetime(dt)

# --- Object parser ---


class Object(FieldSet):
    TYPE_INFO = {
        # TODO: Write parser
        0: ("end[]", None, "End (reserved for BER, None)", None),
        1: ("boolean[]", readBoolean, "Boolean", None),
        2: ("integer[]", readInteger, "Integer", None),
        3: ("bit_str[]", readBitString, "Bit string", None),
        4: ("octet_str[]", readOctetString, "Octet string", None),
        5: ("null[]", None, "NULL (empty, None)", None),
        6: ("obj_id[]", readObjectID, "Object identifier", formatObjectID),
        7: ("obj_desc[]", None, "Object descriptor", None),  # TODO: Write parser
        # TODO: Write parser # External?
        8: ("external[]", None, "External, instance of", None),
        9: ("real[]", readASCIIString, "Real number", None),  # TODO: Write parser
        10: ("enum[]", readInteger, "Enumerated", None),
        11: ("embedded[]", None, "Embedded PDV", None),  # TODO: Write parser
        12: ("utf8_str[]", readUTF8String, "Printable string", None),
        # TODO: Write parser
        13: ("rel_obj_id[]", None, "Relative object identifier", None),
        14: ("time[]", None, "Time", None),  # TODO: Write parser
        # 15: invalid??? sequence of???
        16: ("seq[]", readSequence, "Sequence", None),
        17: ("set[]", readSet, "Set", None),
        18: ("num_str[]", readASCIIString, "Numeric string", None),
        19: ("print_str[]", readASCIIString, "Printable string", formatValue),
        20: ("teletex_str[]", readASCIIString, "Teletex (T61, None) string", None),
        21: ("videotex_str[]", readASCIIString, "Videotex string", None),
        22: ("ia5_str[]", readASCIIString, "IA5 string", formatValue),
        23: ("utc_time[]", readASCIIString, "UTC time", formatUTCTime),
        24: ("general_time[]", readASCIIString, "Generalized time", None),
        25: ("graphic_str[]", readASCIIString, "Graphic string", None),
        26: ("visible_str[]", readASCIIString, "Visible (ISO64, None) string", None),
        27: ("general_str[]", readASCIIString, "General string", None),
        28: ("universal_str[]", readASCIIString, "Universal string", None),
        29: ("unrestricted_str[]", readASCIIString, "Unrestricted string", None),
        30: ("bmp_str[]", readBMPString, "BMP string", None),
        # 31: multiple octet tag number, TODO: not supported

        # Extended tag values:
        #   31: Date
        #   32: Time of day
        #   33: Date-time
        #   34: Duration
    }
    TYPE_DESC = createDict(TYPE_INFO, 2)

    CLASS_DESC = {0: "universal", 1: "application", 2: "context", 3: "private"}
    FORM_DESC = {False: "primitive", True: "constructed"}

    def __init__(self, *args, **kw):
        FieldSet.__init__(self, *args, **kw)
        key = self["type"].value & 31
        if self['class'].value == 0:
            # universal object
            if key in self.TYPE_INFO:
                self._name, self._handler, self._description, create_desc = self.TYPE_INFO[
                    key]
                if create_desc:
                    self.createDescription = lambda: "%s: %s" % (
                        self.TYPE_INFO[key][2], create_desc(self))
                    self._description = None
            elif key == 31:
                raise ParserError(
                    "ASN.1 Object: tag bigger than 30 are not supported")
            else:
                self._handler = None
        elif self['form'].value:
            # constructed: treat as sequence
            self._name = 'seq[]'
            self._handler = readSequence
            self._description = 'constructed object type %i' % key
        else:
            # primitive, context/private
            self._name = 'raw[]'
            self._handler = readASCIIString
            self._description = '%s object type %i' % (
                self['class'].display, key)
        field = self["size"]
        self._size = field.address + field.size + field.value * 8

    def createFields(self):
        yield Enum(Bits(self, "class", 2), self.CLASS_DESC)
        yield Enum(Bit(self, "form"), self.FORM_DESC)
        if self['class'].value == 0:
            yield Enum(Bits(self, "type", 5), self.TYPE_DESC)
        else:
            yield Bits(self, "type", 5)
        yield ASNInteger(self, "size", "Size in bytes")
        size = self["size"].value
        if size:
            if self._handler:
                yield from self._handler(self, size)
            else:
                yield RawBytes(self, "raw", size)


class ASN1File(Parser):
    PARSER_TAGS = {
        "id": "asn1",
        "category": "container",
        "file_ext": ("der",),
        "min_size": 16,
        "description": "Abstract Syntax Notation One (ASN.1)"
    }
    endian = BIG_ENDIAN

    def validate(self):
        try:
            root = self[0]
        except (InputStreamError, FieldError):
            return "Unable to create root object"
        if root.size != self.size:
            return "Invalid root object size"
        return True

    def createFields(self):
        yield Object(self, "root")



"""
ZSNES Save State Parser (v143 only currently)

Author: Jason Gorski
Creation date: 2006-09-15
"""

from hachoir.parser import Parser
from hachoir.field import (FieldSet, StaticFieldSet,
                           UInt8, UInt16, UInt32,
                           String, PaddingBytes, Bytes, RawBytes)
from hachoir.core.endian import LITTLE_ENDIAN


class ZSTHeader(StaticFieldSet):
    format = (
        (String, "zs_mesg", 26, "File header", {"charset": "ASCII"}),
        (UInt8, "zs_mesglen", "File header string len"),
        (UInt8, "zs_version", "Version minor #"),
        (UInt8, "curcyc", "cycles left in scanline"),
        (UInt16, "curypos", "current y position"),
        (UInt8, "cacheud", "update cache every ? frames"),
        (UInt8, "ccud", "current cache increment"),
        (UInt8, "intrset", "interrupt set"),
        (UInt8, "cycpl", "cycles per scanline"),
        (UInt8, "cycphb", "cycles per hblank"),
        (UInt8, "spcon", "SPC Enable (1=enabled)"),
        (UInt16, "stackand", "value to and stack to keep it from going to the wrong area"),
        (UInt16, "stackor", "value to or stack to keep it from going to the wrong area"),
    )


class ZSTcpu(StaticFieldSet):
    format = (
        (UInt16, "xat"),
        (UInt8, "xdbt"),
        (UInt8, "xpbt"),
        (UInt16, "xst"),
        (UInt16, "xdt"),
        (UInt16, "xxt"),
        (UInt16, "xyt"),
        (UInt8, "xp"),
        (UInt8, "xe"),
        (UInt16, "xpc"),
        (UInt8, "xirqb", "which bank the irqs start at"),
        (UInt8, "debugger", "Start with debugger (1: yes, 0: no)"),
        (UInt32, "Curtable" "Current table address"),
        (UInt8, "curnmi", "if in NMI (1=yes)"),
        (UInt32, "cycpbl", "percentage left of CPU/SPC to run (3.58 = 175)"),
        (UInt32, "cycpblt", "percentage of CPU/SPC to run"),
    )


class ZSTppu(FieldSet):
    static_size = 3019 * 8

    def createFields(self):
        yield UInt8(self, "sndrot", "rotates to use A,X or Y for sound skip")
        yield UInt8(self, "sndrot2", "rotates a random value for sound skip")
        yield UInt8(self, "INTEnab", "enables NMI(7)/VIRQ(5)/HIRQ(4)/JOY(0)")
        yield UInt8(self, "NMIEnab", "controlled in e65816 loop. Sets to 81h")
        yield UInt16(self, "VIRQLoc", "VIRQ Y location")
        yield UInt8(self, "vidbright", "screen brightness 0..15")
        yield UInt8(self, "previdbr", "previous screen brightness")
        yield UInt8(self, "forceblnk", "force blanking on/off ($80=on)")
        yield UInt32(self, "objptr", "pointer to object data in VRAM")
        yield UInt32(self, "objptrn", "pointer2 to object data in VRAM")
        yield UInt8(self, "objsize1", "1=8dot, 4=16dot, 16=32dot, 64=64dot")
        yield UInt8(self, "objsize2", "large object size")
        yield UInt8(self, "objmovs1", "number of bytes to move/paragraph")
        yield UInt16(self, "objadds1", "number of bytes to add/paragraph")
        yield UInt8(self, "objmovs2", "number of bytes to move/paragraph")
        yield UInt16(self, "objadds2", "number of bytes to add/paragraph")
        yield UInt16(self, "oamaddrt", "oam address")
        yield UInt16(self, "oamaddrs", "oam address at beginning of vblank")
        yield UInt8(self, "objhipr", "highest priority object #")
        yield UInt8(self, "bgmode", "graphics mode 0..7")
        yield UInt8(self, "bg3highst", "is 1 if background 3 has the highest priority")
        yield UInt8(self, "bgtilesz", "0=8x8, 1=16x16 bit0=bg1, bit1=bg2, etc.")
        yield UInt8(self, "mosaicon", "mosaic on, bit 0=bg1, bit1=bg2, etc.")
        yield UInt8(self, "mosaicsz", "mosaic size in pixels")
        yield UInt16(self, "bg1ptr", "pointer to background1")
        yield UInt16(self, "bg2ptr", "pointer to background2")
        yield UInt16(self, "bg3ptr", "pointer to background3")
        yield UInt16(self, "bg4ptr", "pointer to background4")
        yield UInt16(self, "bg1ptrb", "pointer to background1")
        yield UInt16(self, "bg2ptrb", "pointer to background2")
        yield UInt16(self, "bg3ptrb", "pointer to background3")
        yield UInt16(self, "bg4ptrb", "pointer to background4")
        yield UInt16(self, "bg1ptrc", "pointer to background1")
        yield UInt16(self, "bg2ptrc", "pointer to background2")
        yield UInt16(self, "bg3ptrc", "pointer to background3")
        yield UInt16(self, "bg4ptrc", "pointer to background4")
        yield UInt16(self, "bg1ptrd", "pointer to background1")
        yield UInt16(self, "bg2ptrd", "pointer to background2")
        yield UInt16(self, "bg3ptrd", "pointer to background3")
        yield UInt16(self, "bg4ptrd", "pointer to background4")
        yield UInt8(self, "bg1scsize", "bg #1 screen size (0=1x1,1=1x2,2=2x1,3=2x2)")
        yield UInt8(self, "bg2scsize", "bg #2 screen size (0=1x1,1=1x2,2=2x1,3=2x2)")
        yield UInt8(self, "bg3scsize", "bg #3 screen size (0=1x1,1=1x2,2=2x1,3=2x2)")
        yield UInt8(self, "bg4scsize", "bg #4 screen size (0=1x1,1=1x2,2=2x1,3=2x2)")
        yield UInt16(self, "bg1objptr", "pointer to tiles in background1")
        yield UInt16(self, "bg2objptr", "pointer to tiles in background2")
        yield UInt16(self, "bg3objptr", "pointer to tiles in background3")
        yield UInt16(self, "bg4objptr", "pointer to tiles in background4")
        yield UInt16(self, "bg1scrolx", "background 1 x position")
        yield UInt16(self, "bg2scrolx", "background 2 x position")
        yield UInt16(self, "bg3scrolx", "background 3 x position")
        yield UInt16(self, "bg4scrolx", "background 4 x position")
        yield UInt16(self, "bg1sx", "Temporary Variable for Debugging purposes")
        yield UInt16(self, "bg1scroly", "background 1 y position")
        yield UInt16(self, "bg2scroly", "background 2 y position")
        yield UInt16(self, "bg3scroly", "background 3 y position")
        yield UInt16(self, "bg4scroly", "background 4 y position")
        yield UInt16(self, "addrincr", "vram increment (2,64,128,256)")
        yield UInt8(self, "vramincr", "0 = increment at 2118/2138, 1 = 2119,213A")
        yield UInt8(self, "vramread", "0 = address set, 1 = already read once")
        yield UInt32(self, "vramaddr", "vram address")

        yield UInt16(self, "cgaddr", "cg (palette)")
        yield UInt8(self, "cgmod", "if cgram is modified or not")
        yield UInt16(self, "scrnon", "main & sub screen on")
        yield UInt8(self, "scrndist", "which background is disabled")
        yield UInt16(self, "resolutn", "screen resolution")
        yield UInt8(self, "multa", "multiplier A")
        yield UInt16(self, "diva", "divisor C")
        yield UInt16(self, "divres", "quotent of divc/divb")
        yield UInt16(self, "multres", "result of multa * multb/remainder of divc/divb")
        yield UInt16(self, "latchx", "latched x value")
        yield UInt16(self, "latchy", "latched y value")
        yield UInt8(self, "latchxr", "low or high byte read for x value")
        yield UInt8(self, "latchyr", "low or high byte read for y value")
        yield UInt8(self, "frskipper", "used to control frame skipping")
        yield UInt8(self, "winl1", "window 1 left position")
        yield UInt8(self, "winr1", "window 1 right position")
        yield UInt8(self, "winl2", "window 2 left position")
        yield UInt8(self, "winr2", "window 2 right position")
        yield UInt8(self, "winbg1en", "Win1 on (IN/OUT) or Win2 on (IN/OUT) on BG1")
        yield UInt8(self, "winbg2en", "Win1 on (IN/OUT) or Win2 on (IN/OUT) on BG2")
        yield UInt8(self, "winbg3en", "Win1 on (IN/OUT) or Win2 on (IN/OUT) on BG3")
        yield UInt8(self, "winbg4en", "Win1 on (IN/OUT) or Win2 on (IN/OUT) on BG4")
        yield UInt8(self, "winobjen", "Win1 on (IN/OUT) or Win2 on (IN/OUT) on sprites")
        yield UInt8(self, "wincolen", "Win1 on (IN/OUT) or Win2 on (IN/OUT) on backarea")
        yield UInt8(self, "winlogica", "Window logic type for BG1 to 4")
        yield UInt8(self, "winlogicb", "Window logic type for Sprites and Backarea")
        yield UInt8(self, "winenabm", "Window logic enable for main screen")
        yield UInt8(self, "winenabs", "Window logic enable for sub sceen")
        yield UInt8(self, "mode7set", "mode 7 settings")
        yield UInt16(self, "mode7A", "A value for Mode 7")
        yield UInt16(self, "mode7B", "B value for Mode 7")
        yield UInt16(self, "mode7C", "C value for Mode 7")
        yield UInt16(self, "mode7D", "D value for Mode 7")
        yield UInt16(self, "mode7X0", "Center X for Mode 7")
        yield UInt16(self, "mode7Y0", "Center Y for Mode 7")
        yield UInt8(self, "JoyAPos", "Old-Style Joystick Read Position for Joy 1 & 3")
        yield UInt8(self, "JoyBPos", "Old-Style Joystick Read Position for Joy 2 & 4")
        yield UInt32(self, "compmult", "Complement Multiplication for Mode 7")
        yield UInt8(self, "joyalt", "temporary joystick alternation")
        yield UInt32(self, "wramrwadr", "continuous read/write to wram address")
        yield RawBytes(self, "dmadata", 129, "dma data (written from ports 43xx)")
        yield UInt8(self, "irqon", "if IRQ has been called (80h) or not (0)")
        yield UInt8(self, "nexthdma", "HDMA data to execute once vblank ends")
        yield UInt8(self, "curhdma", "Currently executed hdma")
        yield RawBytes(self, "hdmadata", 152, "4 dword register addresses, # bytes to transfer/line, address increment (word)")
        yield UInt8(self, "hdmatype", "if first time executing hdma or not")
        yield UInt8(self, "coladdr", "red value of color to add")
        yield UInt8(self, "coladdg", "green value of color to add")
        yield UInt8(self, "coladdb", "blue value of color to add")
        yield UInt8(self, "colnull", "keep this 0 (when accessing colors by dword)")
        yield UInt8(self, "scaddset", "screen/fixed color addition settings")
        yield UInt8(self, "scaddtype", "which screen to add/sub")
        yield UInt8(self, "Voice0Disabl2", "Disable Voice 0")
        yield UInt8(self, "Voice1Disabl2", "Disable Voice 1")
        yield UInt8(self, "Voice2Disabl2", "Disable Voice 2")
        yield UInt8(self, "Voice3Disabl2", "Disable Voice 3")
        yield UInt8(self, "Voice4Disabl2", "Disable Voice 4")
        yield UInt8(self, "Voice5Disabl2", "Disable Voice 5")
        yield UInt8(self, "Voice6Disabl2", "Disable Voice 6")
        yield UInt8(self, "Voice7Disabl2", "Disable Voice 7")
        yield RawBytes(self, "oamram", 1024, "OAMRAM (544 bytes)")
        yield RawBytes(self, "cgram", 512, "CGRAM")
        yield RawBytes(self, "pcgram", 512, "Previous CGRAM")
        yield UInt8(self, "vraminctype")
        yield UInt8(self, "vramincby8on", "if increment by 8 is on")
        yield UInt8(self, "vramincby8left", "how many left")
        yield UInt8(self, "vramincby8totl", "how many in total (32,64,128)")
        yield UInt8(self, "vramincby8rowl", "how many left in that row (start at 8)")
        yield UInt16(self, "vramincby8ptri", "increment by how many when rowl = 0")
        yield UInt8(self, "nexthprior")
        yield UInt8(self, "doirqnext")
        yield UInt16(self, "vramincby8var")
        yield UInt8(self, "screstype")
        yield UInt8(self, "extlatch")
        yield UInt8(self, "cfield")
        yield UInt8(self, "interlval")
        yield UInt16(self, "HIRQLoc HIRQ X")

        # NEWer ZST format
        yield UInt8(self, "KeyOnStA")
        yield UInt8(self, "KeyOnStB")
        yield UInt8(self, "SDD1BankA")
        yield UInt8(self, "SDD1BankB")
        yield UInt8(self, "SDD1BankC")
        yield UInt8(self, "SDD1BankD")
        yield UInt8(self, "vramread2")
        yield UInt8(self, "nosprincr")
        yield UInt16(self, "poamaddrs")
        yield UInt8(self, "ioportval")
        yield UInt8(self, "iohvlatch")
        yield UInt8(self, "ppustatus")

        yield PaddingBytes(self, "tempdat", 477, "Reserved/Unused")


class ZSNESFile(Parser):
    PARSER_TAGS = {
        "id": "zsnes",
        "category": "game",
        "description": "ZSNES Save State File (only version 143)",
        "min_size": 3091 * 8,
        "file_ext": ("zst", "zs1", "zs2", "zs3", "zs4", "zs5", "zs6",
                     "zs7", "zs8", "zs9")
    }
    endian = LITTLE_ENDIAN

    def validate(self):
        temp = self.stream.readBytes(0, 28)
        if temp[0:26] != b"ZSNES Save State File V143":
            return "Wrong header"
        if ord(temp[27:28]) != 143:  # extra...
            return "Wrong save version %d <> 143" % temp[27:1]
        return True

    def seek(self, offset):
        padding = self.seekByte(offset, relative=False)
        if padding is not None:
            yield padding

    def createFields(self):
        yield ZSTHeader(self, "header", "ZST header")  # Offset: 0
        yield ZSTcpu(self, "cpu", "ZST cpu registers")  # 41
        yield ZSTppu(self, "ppu", "ZST CPU registers")  # 72
        yield RawBytes(self, "wram7E", 65536)  # 3091
        yield RawBytes(self, "wram7F", 65536)  # 68627
        yield RawBytes(self, "vram", 65536)  # 134163

        # TODO: Interpret extra on-cart chip data found at/beyond... 199699

        # TODO: Interpret Thumbnail/Screenshot data found at 275291
        # 64*56*2(16bit colors) = 7168
        padding = self.seekByte(275291, relative=False)
        if padding is not None:
            yield padding
        yield Bytes(self, "thumbnail", 7168, "Thumbnail of playing game in some sort of raw 64x56x16-bit RGB mode?")



"""
Canon CR2 raw image data, version 2.0 image parser.

Authors: Fernando Crespo
Creation date: 21 february 2017
"""

from hachoir.parser import Parser
from hachoir.field import SeekableFieldSet, RootSeekableFieldSet, Bytes, String, UInt8, UInt16, UInt32
from hachoir.core.endian import LITTLE_ENDIAN, BIG_ENDIAN
from hachoir.core.text_handler import textHandler, hexadecimal
from hachoir.parser.image.exif import IFD, IFD_TAGS


def getStrips(ifd):
    data = {}
    for i, entry in enumerate(ifd.array('entry')):
        data[entry['tag'].display] = entry
    # image data
    if "StripOffsets" in data and "StripByteCounts" in data:
        offs = ifd.getEntryValues(data["StripOffsets"])
        bytes = ifd.getEntryValues(data["StripByteCounts"])
        for off, byte in zip(offs, bytes):
            yield off.value, byte.value


class ImageFile(SeekableFieldSet):

    def __init__(self, parent, name, description, ifd):
        SeekableFieldSet.__init__(self, parent, name, description, None)
        self._ifd = ifd

    def createFields(self):
        for off, byte in getStrips(self._ifd):
            self.seekByte(off, relative=False)
            yield Bytes(self, "strip[]", byte)


class CR2File(RootSeekableFieldSet, Parser):
    PARSER_TAGS = {
        "id": "cr2",
        "category": "image",
        "file_ext": ("cr2",),
        "mime": ("image/x-canon-cr2",),
        "min_size": 15,
        "magic": ((b"CR", 8),),
        "description": "Canon CR2 raw image data, version 2.0"
    }

    # Correct endian is set in constructor
    endian = LITTLE_ENDIAN

    def __init__(self, stream, **args):
        RootSeekableFieldSet.__init__(
            self, None, "root", stream, None, stream.askSize(self))
        if self.stream.readBytes(0, 2) == b"MM":
            self.endian = BIG_ENDIAN
        Parser.__init__(self, stream, **args)

    def validate(self):
        endian = self.stream.readBytes(0, 2)
        if endian not in (b"MM", b"II"):
            return "Invalid endian (%r)" % endian
        if self["version"].value != 42:
            return "Unknown Canon TIFF version - " + str(self["version"].value)
        if self["cr_identifier"].value != "CR":
            return "Unknown Canon Raw File"
        return True

    def createFields(self):
        iff_start = self.absolute_address
        yield String(self, "endian", 2, "Endian ('II' or 'MM')", charset="ASCII")
        if self["endian"].value == "II":
            self.endian = LITTLE_ENDIAN
        else:
            self.endian = BIG_ENDIAN

        yield UInt16(self, "version", "TIFF version number")
        yield UInt32(self, "img_dir_ofs", "Next image directory offset")

        yield String(self, "cr_identifier", 2, "Canon Raw marker", charset="ASCII")
        yield UInt8(self, "cr_major_version", "Canon Raw major version number")
        yield UInt8(self, "cr_minor_version", "Canon Raw minor version number")

        yield textHandler(UInt32(self, "cr_raw_ifd_offset", "Offset to Raw IFD"), hexadecimal)

        offsets = [(self['img_dir_ofs'].value, 'ifd[]', IFD)]

        while offsets:
            offset, name, klass = offsets.pop(0)
            self.seekByte(offset + iff_start // 8, relative=False)
            ifd = klass(self, name, iff_start)

            yield ifd
            for entry in ifd.array('entry'):
                tag = entry['tag'].value
                if tag in IFD_TAGS:
                    name, klass = IFD_TAGS[tag]
                    offsets.append((ifd.getEntryValues(entry)[
                                   0].value, name + '[]', klass))
            if ifd['next'].value != 0:
                offsets.append((ifd['next'].value, 'ifd[]', IFD))

        for ifd in self.array('ifd'):
            offs = (off for off, byte in getStrips(ifd))
            self.seekByte(min(offs), relative=False)
            image = ImageFile(self, "image[]", "Image File", ifd)
            yield image



"""
ReiserFS file system version 3 parser (other version have not been tested).

Author: Frederic Weisbecker
Creation date: 8 december 2006

Sources:
 - http://p-nand-q.com/download/rfstool/reiserfs_docs.html
 - http://homes.cerias.purdue.edu/~florian/reiser/reiserfs.php
 - file://usr/src/linux-2.6.16.19/include/linux/reiserfs_fs.h

NOTES:

The most part of the description of the structures, their fields and their
comments decribed here comes from the file include/linux/reiserfs_fs.h
- written by Hans reiser - located in the Linux kernel 2.6.16.19 and from
the Reiserfs explanations in
http://p-nand-q.com/download/rfstool/reiserfs_docs.html written by Gerson
Kurz.
"""


from hachoir.parser import Parser
from hachoir.field import (FieldSet, Enum,
                           UInt16, UInt32, String, RawBytes, NullBytes, SeekableFieldSet, Bit)
from hachoir.core.endian import LITTLE_ENDIAN


class BlockState(Bit):
    """The state (used/free) of a ReiserFs Block"""

    STATE = {
        True: "used",
        False: "free"
    }

    block_nb = 0

    def __init__(self, parent, name, nb_block):
        """@param nb_block: Number of the block concerned"""
        Bit.__init__(self, parent, name)
        self.block_nb = self.__class__.block_nb
        self.__class__.block_nb += 1

    def createDescription(self):
        return "State of the block %d" % self.block_nb

    def createDisplay(self):
        return self.STATE[Bit.createValue(self)]


class BitmapBlock(SeekableFieldSet):
    """ The bitmap blocks are Reiserfs blocks where each byte contains
        the state of 8 blocks in the filesystem. So each bit will describe
        the state of a block to tell if it is used or not.
    """

    def createFields(self):
        block_size = self["/superblock/blocksize"].value

        for i in range(0, block_size * 8):
            yield BlockState(self, "block[]", i)


class BitmapBlockGroup(SeekableFieldSet):
    """The group that manages the Bitmap Blocks"""

    def createFields(self):
        block_size = self["/superblock/blocksize"].value
        nb_bitmap_block = self["/superblock/bmap_nr"].value
        # Position of the first bitmap block
        self.seekByte(REISER_FS.SUPERBLOCK_OFFSET + block_size, relative=False)

        yield BitmapBlock(self, "BitmapBlock[]", "Bitmap blocks tells for each block if it is used")
        # The other bitmap blocks
        for i in range(1, nb_bitmap_block):
            self.seekByte((block_size**2) * 8 * i, relative=False)
            yield BitmapBlock(self, "BitmapBlock[]", "Bitmap blocks tells for each block if it is used")


class Journal_params(FieldSet):
    static_size = 32 * 8

    def createFields(self):
        yield UInt32(self, "1st_block", "Journal 1st block number")
        yield UInt32(self, "dev", "Journal device number")
        yield UInt32(self, "size", "Size of the journal")
        yield UInt32(self, "trans_max", "Max number of blocks in a transaction")
        # TODO: Must be explained: it was sb_journal_block_count
        yield UInt32(self, "magic", "Random value made on fs creation.")
        yield UInt32(self, "max_batch", "Max number of blocks to batch into a trans")
        yield UInt32(self, "max_commit_age", "In seconds, how old can an async commit be")
        yield UInt32(self, "max_trans_age", "In seconds, how old can a transaction be")

    def createDescription(self):
        return "Parameters of the journal"


class SuperBlock(FieldSet):
    # static_size = 204*8

    UMOUNT_STATE = {1: "unmounted", 2: "not unmounted"}
    HASH_FUNCTIONS = {
        0: "UNSET_HASH",
        1: "TEA_HASH",
        2: "YURA_HASH",
        3: "R5_HASH"
    }

    def createFields(self):
        # TODO: This structure is normally divided in two parts:
        # _reiserfs_super_block_v1
        # _reiserfs_super_block
        # It will be divided later to easily support older version of the first
        # part
        yield UInt32(self, "block_count", "Number of blocks")
        yield UInt32(self, "free_blocks", "Number of free blocks")
        yield UInt32(self, "root_block", "Root block number")
        yield Journal_params(self, "Journal parameters")
        yield UInt16(self, "blocksize", "Size of a block")
        yield UInt16(self, "oid_maxsize", "Max size of object id array")
        yield UInt16(self, "oid_cursize", "Current size of object id array")
        yield Enum(UInt16(self, "umount_state", "Filesystem umounted or not"), self.UMOUNT_STATE)
        yield String(self, "magic", 10, "Magic string", strip="\0")
        # TODO: change the type of s_fs_state in Enum to have more details
        # about this fsck state
        yield UInt16(self, "fs_state", "Rebuilding phase of fsck ")
        yield Enum(UInt32(self, "hash_function", "Hash function to sort names in a directory"), self.HASH_FUNCTIONS)
        yield UInt16(self, "tree_height", "Height of disk tree")
        yield UInt16(self, "bmap_nr", "Amount of bitmap blocks needed to address each block of file system")
        # TODO: find a good description for this field
        yield UInt16(self, "version", "Field only reliable on filesystem with non-standard journal")
        yield UInt16(self, "reserved_for_journal", "Size in blocks of journal area on main device")
        # TODO: same as above
        yield UInt32(self, "inode_generation", "No description")
        # TODO: same as above and should be an enum field
        yield UInt32(self, "flags", "No description")
        # TODO: Create a special Type to format this id
        yield RawBytes(self, "uuid", 16, "Filesystem unique identifier")
        yield String(self, "label", 16, "Filesystem volume label", strip="\0")
        yield NullBytes(self, "unused", 88)
        yield NullBytes(self, "Bytes before end of the block", self["blocksize"].value - 204)

    def createDescription(self):
        return "Superblock: ReiserFs Filesystem"


class REISER_FS(Parser):
    PARSER_TAGS = {
        "id": "reiserfs",
        "category": "file_system",
        # 130 blocks before the journal +
        # Minimal size of journal (513 blocks) +
        # 1 block for the rest
        # And The Minimal size of a block is 512 bytes
        "min_size": (130 + 513 + 1) * (512 * 8),
        "description": "ReiserFS file system"
    }
    endian = LITTLE_ENDIAN

    # Offsets (in bytes) of important information
    SUPERBLOCK_OFFSET = 64 * 1024
    MAGIC_OFFSET = SUPERBLOCK_OFFSET + 52

    def validate(self):
        # Let's look at the magic field in the superblock
        magic = self.stream.readBytes(self.MAGIC_OFFSET * 8, 9).rstrip(b"\0")
        if magic in (b"ReIsEr3Fs", b"ReIsErFs", b"ReIsEr2Fs"):
            return True
        return "Invalid magic string"

    def createFields(self):
        yield NullBytes(self, "padding[]", self.SUPERBLOCK_OFFSET)
        yield SuperBlock(self, "superblock")
        yield BitmapBlockGroup(self, "Group of bitmap blocks")



"""
Character field class: a 8-bit character
"""

from hachoir.field import Bits
from hachoir.core.tools import makePrintable


class Character(Bits):
    """
    A 8-bit character using ASCII charset for display attribute.
    """
    static_size = 8

    def __init__(self, parent, name, description=None):
        Bits.__init__(self, parent, name, self.static_size, description=description)

    def createValue(self):
        return chr(self._parent.stream.readBits(
            self.absolute_address, self.static_size, self.parent.endian))

    def createRawDisplay(self):
        return str(Bits.createValue(self))

    def createDisplay(self):
        return makePrintable(self.value, "ASCII", quote="'")



"""
TrueType Font parser.

Documents:
 - "The OpenType Specification"
   https://docs.microsoft.com/en-us/typography/opentype/spec/
 - "An Introduction to TrueType Fonts: A look inside the TTF format"
   written by "NRSI: Computers & Writing Systems"
   http://scripts.sil.org/cms/scripts/page.php?site_id=nrsi&item_id=IWS-Chapter08

Author: Victor Stinner
Creation date: 2007-02-08
"""

from hachoir.parser import Parser
from hachoir.field import (
    FieldSet,
    ParserError,
    UInt8,
    UInt16,
    UInt24,
    UInt32,
    Int16,
    Bit,
    Bits,
    PaddingBits,
    NullBytes,
    String,
    RawBytes,
    Bytes,
    Enum,
    TimestampMac32,
    GenericVector,
    PascalString8,
)
from hachoir.core.endian import BIG_ENDIAN
from hachoir.core.text_handler import textHandler, hexadecimal, filesizeHandler

MAX_NAME_COUNT = 300
MIN_NB_TABLE = 3
MAX_NB_TABLE = 30

DIRECTION_NAME = {
    0: "Mixed directional",
    1: "Left to right",
    2: "Left to right + neutrals",
    -1: "Right to left",
    -2: "Right to left + neutrals",
}

NAMEID_NAME = {
    0: "Copyright notice",
    1: "Font family name",
    2: "Font subfamily name",
    3: "Unique font identifier",
    4: "Full font name",
    5: "Version string",
    6: "Postscript name",
    7: "Trademark",
    8: "Manufacturer name",
    9: "Designer",
    10: "Description",
    11: "URL Vendor",
    12: "URL Designer",
    13: "License Description",
    14: "License info URL",
    16: "Preferred Family",
    17: "Preferred Subfamily",
    18: "Compatible Full",
    19: "Sample text",
    20: "PostScript CID findfont name",
}

PLATFORM_NAME = {
    0: "Unicode",
    1: "Macintosh",
    2: "ISO",
    3: "Microsoft",
    4: "Custom",
}

CHARSET_MAP = {
    # (platform, encoding) => charset
    0: {3: "UTF-16-BE"},
    1: {0: "MacRoman"},
    3: {1: "UTF-16-BE"},
}

PERMISSIONS = {
    0: "Installable embedding",
    2: "Restricted License embedding",
    4: "Preview & Print embedding",
    8: "Editable embedding",
}

FWORD = Int16
UFWORD = UInt16


class Tag(String):
    def __init__(self, parent, name, description=None):
        String.__init__(self, parent, name, 4, description)


class Version16Dot16(FieldSet):
    static_size = 32

    def createFields(self):
        yield UInt16(self, "major")
        yield UInt16(self, "minor")

    def createValue(self):
        return float("%u.%x" % (self["major"].value, self["minor"].value))


class Fixed(FieldSet):
    def createFields(self):
        yield UInt16(self, "int_part")
        yield UInt16(self, "float_part")

    def createValue(self):
        return self["int_part"].value + float(self["float_part"].value) / 65536


class Tuple(FieldSet):
    def __init__(self, parent, name, axisCount):
        super().__init__(parent, name, description="Tuple Record")
        self.axisCount = axisCount

    def createFields(self):
        for _ in range(self.axisCount):
            yield (Fixed(self, "coordinate[]"))


class F2DOT14(FieldSet):
    static_size = 16

    def createFields(self):
        yield Int16(self, "int_part")

    def createValue(self):
        return self["int_part"].value / 16384


class TableHeader(FieldSet):
    def createFields(self):
        yield Tag(self, "tag")
        yield textHandler(UInt32(self, "checksum"), hexadecimal)
        yield UInt32(self, "offset")
        yield filesizeHandler(UInt32(self, "size"))

    def createDescription(self):
        return "Table entry: %s (%s)" % (self["tag"].display, self["size"].display)


class NameHeader(FieldSet):
    def createFields(self):
        yield Enum(UInt16(self, "platformID"), PLATFORM_NAME)
        yield UInt16(self, "encodingID")
        yield UInt16(self, "languageID")
        yield Enum(UInt16(self, "nameID"), NAMEID_NAME)
        yield UInt16(self, "length")
        yield UInt16(self, "offset")

    def getCharset(self):
        platform = self["platformID"].value
        encoding = self["encodingID"].value
        try:
            return CHARSET_MAP[platform][encoding]
        except KeyError:
            self.warning("TTF: Unknown charset (%s,%s)" % (platform, encoding))
            return "ISO-8859-1"

    def createDescription(self):
        platform = self["platformID"].display
        name = self["nameID"].display
        return "Name record: %s (%s)" % (name, platform)


def parseFontHeader(self):
    yield UInt16(self, "maj_ver", "Major version")
    yield UInt16(self, "min_ver", "Minor version")
    yield UInt16(self, "font_maj_ver", "Font major version")
    yield UInt16(self, "font_min_ver", "Font minor version")
    yield textHandler(UInt32(self, "checksum"), hexadecimal)
    yield Bytes(self, "magic", 4, r"Magic string (\x5F\x0F\x3C\xF5)")
    if self["magic"].value != b"\x5F\x0F\x3C\xF5":
        raise ParserError("TTF: invalid magic of font header")

    # Flags
    yield Bit(self, "y0", "Baseline at y=0")
    yield Bit(self, "x0", "Left sidebearing point at x=0")
    yield Bit(self, "instr_point", "Instructions may depend on point size")
    yield Bit(self, "ppem", "Force PPEM to integer values for all")
    yield Bit(self, "instr_width", "Instructions may alter advance width")
    yield Bit(self, "vertical", "e laid out vertically?")
    yield PaddingBits(self, "reserved[]", 1)
    yield Bit(self, "linguistic", "Requires layout for correct linguistic rendering?")
    yield Bit(self, "gx", "Metamorphosis effects?")
    yield Bit(self, "strong", "Contains strong right-to-left glyphs?")
    yield Bit(self, "indic", "contains Indic-style rearrangement effects?")
    yield Bit(self, "lossless", "Data is lossless (Agfa MicroType compression)")
    yield Bit(self, "converted", "Font converted (produce compatible metrics)")
    yield Bit(self, "cleartype", "Optimised for ClearType")
    yield Bits(self, "adobe", 2, "(used by Adobe)")

    yield UInt16(self, "unit_per_em", "Units per em")
    if not (16 <= self["unit_per_em"].value <= 16384):
        raise ParserError("TTF: Invalid unit/em value")
    yield UInt32(self, "created_high")
    yield TimestampMac32(self, "created")
    yield UInt32(self, "modified_high")
    yield TimestampMac32(self, "modified")
    yield UInt16(self, "xmin")
    yield UInt16(self, "ymin")
    yield UInt16(self, "xmax")
    yield UInt16(self, "ymax")

    # Mac style
    yield Bit(self, "bold")
    yield Bit(self, "italic")
    yield Bit(self, "underline")
    yield Bit(self, "outline")
    yield Bit(self, "shadow")
    yield Bit(self, "condensed", "(narrow)")
    yield Bit(self, "expanded")
    yield PaddingBits(self, "reserved[]", 9)

    yield UInt16(self, "lowest", "Smallest readable size in pixels")
    yield Enum(UInt16(self, "font_dir", "Font direction hint"), DIRECTION_NAME)
    yield Enum(UInt16(self, "ofst_format"), {0: "short offsets", 1: "long"})
    yield UInt16(self, "glyph_format", "(=0)")


class AxisValueMap(FieldSet):
    static_size = 32

    def createFields(self):
        yield F2DOT14(self, "fromCoordinate")
        yield F2DOT14(self, "toCoordinate")


class SegmentMaps(FieldSet):
    def createFields(self):
        yield UInt16(
            self, "positionMapCount", "The number of correspondence pairs for this axis"
        )
        for _ in range(self["positionMapCount"].value):
            yield (AxisValueMap(self, "axisValueMaps[]"))


def parseAvar(self):
    yield UInt16(self, "majorVersion", "Major version")
    yield UInt16(self, "minorVersion", "Minor version")
    yield PaddingBits(self, "reserved[]", 16)
    yield UInt16(self, "axisCount", "The number of variation axes for this font")
    for _ in range(self["axisCount"].value):
        yield (SegmentMaps(self, "segmentMaps[]"))


class VariationAxisRecord(FieldSet):
    def createFields(self):
        yield Tag(self, "axisTag", "Tag identifying the design variation for the axis")
        yield Fixed(self, "minValue", "The minimum coordinate value for the axis")
        yield Fixed(self, "defaultValue", "The default coordinate value for the axis")
        yield Fixed(self, "maxValue", "The maximum coordinate value for the axis")
        yield PaddingBits(self, "reservedFlags", 15)
        yield Bit(
            self, "hidden", "The axis should not be exposed directly in user interfaces"
        )
        yield UInt16(
            self,
            "axisNameID",
            "The name ID for entries in the 'name' table that provide a display name for this axis",
        )


class InstanceRecord(FieldSet):
    def __init__(self, parent, name, axisCount, hasPSNameID=False):
        super().__init__(parent, name, description="Instance record")
        self.axisCount = axisCount
        self.hasPSNameID = hasPSNameID

    def createFields(self):
        yield UInt16(
            self, "subfamilyNameID", "Name ID for subfamily names for this instance"
        )
        yield PaddingBits(self, "reservedFlags", 16)
        yield Tuple(self, "coordinates", axisCount=self.axisCount)
        if self.hasPSNameID:
            yield UInt16(
                self,
                "postScriptNameID",
                "Name ID for PostScript names for this instance",
            )


def parseFvar(self):
    yield UInt16(self, "majorVersion", "Major version")
    yield UInt16(self, "minorVersion", "Minor version")
    yield UInt16(
        self, "axisArrayOffset", "Offset to the start of the VariationAxisRecord array."
    )
    yield PaddingBits(self, "reserved[]", 16)
    yield UInt16(self, "axisCount", "The number of variation axes for this font")
    yield UInt16(self, "axisSize", "The size in bytes of each VariationAxisRecord")
    yield UInt16(self, "instanceCount", "The number of named instances for this font")
    yield UInt16(self, "instanceSize", "The size in bytes of each InstanceRecord")
    if self["axisArrayOffset"].value > 16:
        yield PaddingBits(self, "padding", 8 * (self["axisArrayOffset"].value - 16))
    for _ in range(self["axisCount"].value):
        yield (VariationAxisRecord(self, "axes[]"))
    for _ in range(self["instanceCount"].value):
        yield (
            InstanceRecord(
                self,
                "instances[]",
                axisCount=self["axisCount"].value,
                hasPSNameID=(
                    self["instanceSize"].value == (2 * self["axisCount"].value + 6)
                ),
            )
        )


class EncodingRecord(FieldSet):
    static_size = 64

    def createFields(self):
        yield Enum(UInt16(self, "platformID"), PLATFORM_NAME)
        yield UInt16(self, "encodingID")
        self.offset = UInt32(self, "subtableOffset")
        yield self.offset


class CmapTable0(FieldSet):
    def createFields(self):
        yield UInt16(self, "format", "Table format")
        yield UInt16(self, "length", "Length in bytes")
        yield UInt16(self, "language", "Language ID")
        yield GenericVector(self, "mapping", 256, UInt8)


class CmapTable4(FieldSet):
    def createFields(self):
        yield UInt16(self, "format", "Table format")
        yield UInt16(self, "length", "Length in bytes")
        yield UInt16(self, "language", "Language ID")
        yield UInt16(self, "segCountX2", "Twice the number of segments")
        segments = self["segCountX2"].value // 2
        yield UInt16(self, "searchRange")
        yield UInt16(self, "entrySelector")
        yield UInt16(self, "rangeShift")
        yield GenericVector(self, "endCode", segments, UInt16)
        yield PaddingBits(self, "reserved[]", 16)
        yield GenericVector(self, "startCode", segments, UInt16)
        yield GenericVector(self, "idDelta", segments, Int16)
        yield GenericVector(self, "idRangeOffsets", segments, UInt16)
        remainder = (self["length"].value - (self.current_size / 8)) / 2
        if remainder:
            yield GenericVector(self, "glyphIdArray", remainder, UInt16)


class CmapTable6(FieldSet):
    def createFields(self):
        yield UInt16(self, "format", "Table format")
        yield UInt16(self, "length", "Length in bytes")
        yield UInt16(self, "language", "Language ID")
        yield UInt16(self, "firstCode", "First character code of subrange")
        yield UInt16(self, "entryCount", "Number of character codes in subrange")
        yield GenericVector(self, "glyphIdArray", self["entryCount"].value, UInt16)


class SequentialMapGroup(FieldSet):
    def createFields(self):
        yield UInt32(self, "startCharCode", "First character code in this group")
        yield UInt32(self, "endCharCode", "First character code in this group")
        yield UInt32(
            self,
            "startGlyphID",
            "Glyph index corresponding to the starting character code",
        )


class CmapTable12(FieldSet):
    def createFields(self):
        yield UInt16(self, "format", "Table format")
        yield PaddingBits(self, "reserved[]", 16)
        yield UInt32(self, "length", "Length in bytes")
        yield UInt32(self, "language", "Language ID")
        yield UInt32(self, "numGroups", "Number of groupings which follow")
        for i in range(self["numGroups"].value):
            yield SequentialMapGroup(self, "mapgroup[]")


class VariationSelector(FieldSet):
    def createFields(self):
        yield UInt24(self, "varSelector", "Variation selector")
        yield UInt32(self, "defaultUVSOffset", "Offset to default UVS table")
        yield UInt32(self, "nonDefaultUVSOffset", "Offset to non-default UVS table")


class CmapTable14(FieldSet):
    def createFields(self):
        yield UInt16(self, "format", "Table format")
        yield UInt32(self, "length", "Length in bytes")
        yield UInt32(
            self, "numVarSelectorRecords", "Number of variation selector records"
        )
        for i in range(self["numVarSelectorRecords"].value):
            yield VariationSelector(self, "variationSelector[]")


def parseCmap(self):
    yield UInt16(self, "version")
    numTables = UInt16(self, "numTables", "Number of encoding tables")
    yield numTables
    encodingRecords = []
    for index in range(numTables.value):
        entry = EncodingRecord(self, "encodingRecords[]")
        yield entry
        encodingRecords.append(entry)
    encodingRecords.sort(key=lambda field: field["subtableOffset"].value)
    last = None
    for er in encodingRecords:
        offset = er["subtableOffset"].value
        if last and last == offset:
            continue
        last = offset

        # Add padding if any
        padding = self.seekByte(offset, relative=True, null=False)
        if padding:
            yield padding
        format = UInt16(self, "format").value
        if format == 0:
            yield CmapTable0(self, "cmap table format 0")
        elif format == 4:
            yield CmapTable4(self, "cmap table format 4")
        elif format == 6:
            yield CmapTable6(self, "cmap table format 6")
        elif format == 12:
            yield CmapTable12(self, "cmap table format 12")
        elif format == 14:
            yield CmapTable14(self, "cmap table format 14")


class SignatureRecord(FieldSet):
    def createFields(self):
        yield UInt16(self, "format", "Table format")
        yield UInt16(self, "length", "Length of signature")
        yield UInt16(self, "signatureBlockOffset", "Offset to signature block")


class SignatureBlock(FieldSet):
    def createFields(self):
        yield PaddingBits(self, "reserved[]", 32)
        yield UInt32(
            self,
            "length",
            "Length (in bytes) of the PKCS#7 packet in the signature field",
        )
        yield String(self, "signature", self["length"].value, "Signature block")


def parseDSIG(self):
    yield UInt32(self, "version")
    yield UInt16(self, "numSignatures", "Number of signatures in the table")
    yield Bit(self, "flag", "Cannot be resigned")
    yield PaddingBits(self, "reserved[]", 7)
    entries = []
    for i in range(self["numSignatures"].value):
        record = SignatureRecord(self, "signatureRecords[]")
        entries.append(record)
        yield record
    entries.sort(key=lambda field: field["signatureBlockOffset"].value)
    last = None
    for entry in entries:
        offset = entry["signatureBlockOffset"].value
        if last and last == offset:
            continue
        last = offset
        # Add padding if any
        padding = self.seekByte(offset, relative=True, null=False)
        if padding:
            yield padding

    padding = (self.size - self.current_size) // 8
    if padding:
        yield NullBytes(self, "padding_end", padding)


def parseNames(self):
    # Read header
    yield UInt16(self, "format")
    if self["format"].value != 0:
        raise ParserError("TTF (names): Invalid format (%u)" % self["format"].value)
    yield UInt16(self, "count")
    yield UInt16(self, "offset")
    if MAX_NAME_COUNT < self["count"].value:
        raise ParserError("Invalid number of names (%s)" % self["count"].value)

    # Read name index
    entries = []
    for index in range(self["count"].value):
        entry = NameHeader(self, "header[]")
        yield entry
        entries.append(entry)

    # Sort names by their offset
    entries.sort(key=lambda field: field["offset"].value)

    # Read name value
    last = None
    for entry in entries:
        # Skip duplicates values
        new = (entry["offset"].value, entry["length"].value)
        if last and last == new:
            self.warning("Skip duplicate %s %s" % (entry.name, new))
            continue
        last = (entry["offset"].value, entry["length"].value)

        # Skip negative offset
        offset = entry["offset"].value + self["offset"].value
        if offset < self.current_size // 8:
            self.warning("Skip value %s (negative offset)" % entry.name)
            continue

        # Add padding if any
        padding = self.seekByte(offset, relative=True, null=True)
        if padding:
            yield padding

        # Read value
        size = entry["length"].value
        if size:
            yield String(
                self, "value[]", size, entry.description, charset=entry.getCharset()
            )

    padding = (self.size - self.current_size) // 8
    if padding:
        yield NullBytes(self, "padding_end", padding)


def parseMaxp(self):
    # Read header
    yield Version16Dot16(self, "format", "format version")
    yield UInt16(self, "numGlyphs", "Number of glyphs")
    if self["format"].value >= 1:
        yield UInt16(self, "maxPoints", "Maximum points in a non-composite glyph")
        yield UInt16(self, "maxContours", "Maximum contours in a non-composite glyph")
        yield UInt16(self, "maxCompositePoints", "Maximum points in a composite glyph")
        yield UInt16(
            self, "maxCompositeContours", "Maximum contours in a composite glyph"
        )
        yield UInt16(self, "maxZones", "Do instructions use the twilight zone?")
        yield UInt16(self, "maxTwilightPoints", "Maximum points used in Z0")
        yield UInt16(self, "maxStorage", "Number of Storage Area locations")
        yield UInt16(self, "maxFunctionDefs", "Number of function definitions")
        yield UInt16(self, "maxInstructionDefs", "Number of instruction definitions")
        yield UInt16(self, "maxStackElements", "Maximum stack depth")
        yield UInt16(
            self, "maxSizeOfInstructions", "Maximum byte count for glyph instructions"
        )
        yield UInt16(
            self,
            "maxComponentElements",
            "Maximum number of components at glyph top level",
        )
        yield UInt16(self, "maxComponentDepth", "Maximum level of recursion")


def parseHhea(self):
    yield UInt16(self, "majorVersion", "Major version")
    yield UInt16(self, "minorVersion", "Minor version")
    yield FWORD(self, "ascender", "Typographic ascent")
    yield FWORD(self, "descender", "Typographic descent")
    yield FWORD(self, "lineGap", "Typographic linegap")
    yield UFWORD(self, "advanceWidthMax", "Maximum advance width")
    yield FWORD(self, "minLeftSideBearing", "Minimum left sidebearing value")
    yield FWORD(self, "minRightSideBearing", "Minimum right sidebearing value")
    yield FWORD(self, "xMaxExtent", "Maximum X extent")
    yield Int16(self, "caretSlopeRise", "Caret slope rise")
    yield Int16(self, "caretSlopeRun", "Caret slope run")
    yield Int16(self, "caretOffset", "Caret offset")
    yield GenericVector(self, "reserved", 4, Int16)
    yield Int16(self, "metricDataFormat", "Metric data format")
    yield UInt16(self, "numberOfHMetrics", "Number of horizontal metrics")


class fsType(FieldSet):
    def createFields(self):
        yield Enum(Bits(self, "usage_permissions", 4), PERMISSIONS)
        yield PaddingBits(self, "reserved[]", 4)
        yield Bit(self, "no_subsetting", "Font may not be subsetted prior to embedding")
        yield Bit(
            self,
            "bitmap_embedding",
            "Only bitmaps contained in the font may be embedded",
        )
        yield PaddingBits(self, "reserved[]", 6)


def parseOS2(self):
    yield UInt16(self, "version", "Table version")
    yield Int16(self, "xAvgCharWidth")
    yield UInt16(self, "usWeightClass")
    yield UInt16(self, "usWidthClass")
    yield fsType(self, "fsType")
    yield Int16(self, "ySubscriptXSize")
    yield Int16(self, "ySubscriptYSize")
    yield Int16(self, "ySubscriptXOffset")
    yield Int16(self, "ySubscriptYOffset")
    yield Int16(self, "ySuperscriptXSize")
    yield Int16(self, "ySuperscriptYSize")
    yield Int16(self, "ySuperscriptXOffset")
    yield Int16(self, "ySuperscriptYOffset")
    yield Int16(self, "yStrikeoutSize")
    yield Int16(self, "yStrikeoutPosition")
    yield Int16(self, "sFamilyClass")
    yield GenericVector(self, "panose", 10, UInt8)
    yield UInt32(self, "ulUnicodeRange1")
    yield UInt32(self, "ulUnicodeRange2")
    yield UInt32(self, "ulUnicodeRange3")
    yield UInt32(self, "ulUnicodeRange4")
    yield Tag(self, "achVendID", "Vendor ID")
    yield UInt16(self, "fsSelection")
    yield UInt16(self, "usFirstCharIndex")
    yield UInt16(self, "usLastCharIndex")
    yield Int16(self, "sTypoAscender")
    yield Int16(self, "sTypoDescender")
    yield Int16(self, "sTypoLineGap")
    yield UInt16(self, "usWinAscent")
    yield UInt16(self, "usWinDescent")
    if self["version"].value >= 1:
        yield UInt32(self, "ulCodePageRange1")
        yield UInt32(self, "ulCodePageRange2")
    if self["version"].value >= 2:
        yield Int16(self, "sxHeight")
        yield Int16(self, "sCapHeight")
        yield UInt16(self, "usDefaultChar")
        yield UInt16(self, "usBreakChar")
        yield UInt16(self, "usMaxContext")
    if self["version"].value >= 5:
        yield UInt16(self, "usLowerOpticalPointSize")
        yield UInt16(self, "usUpperOpticalPointSize")


def parsePost(self):
    yield Version16Dot16(self, "version", "Table version")
    yield Fixed(
        self,
        "italicAngle",
        "Italic angle in counter-clockwise degrees from the vertical.",
    )
    yield FWORD(self, "underlinePosition", "Top of underline to baseline")
    yield FWORD(self, "underlineThickness", "Suggested underline thickness")
    yield UInt32(self, "isFixedPitch", "Is the font fixed pitch?")
    yield UInt32(self, "minMemType42", "Minimum memory usage (OpenType)")
    yield UInt32(self, "maxMemType42", "Maximum memory usage (OpenType)")
    yield UInt32(self, "minMemType1", "Minimum memory usage (Type 1)")
    yield UInt32(self, "maxMemType1", "Maximum memory usage (Type 1)")
    if self["version"].value == 2.0:
        yield UInt16(self, "numGlyphs")
        indices = GenericVector(
            self,
            "Array of indices into the string data",
            self["numGlyphs"].value,
            UInt16,
            "glyphNameIndex",
        )
        yield indices
        for gid, index in enumerate(indices):
            if index.value >= 258:
                yield PascalString8(self, "glyphname[%i]" % gid)
    elif self["version"].value == 2.0:
        yield UInt16(self, "numGlyphs")
        indices = GenericVector(
            self,
            "Difference between graphic index and standard order of glyph",
            self["numGlyphs"].value,
            UInt16,
            "offset",
        )
        yield indices


# This is work-in-progress until I work out good ways to do random-access on offsets
parseScriptList = (
    parseFeatureList
) = parseLookupList = parseFeatureVariationsTable = lambda x: None


def parseGSUB(self):
    yield UInt16(self, "majorVersion", "Major version")
    yield UInt16(self, "minorVersion", "Minor version")
    SUBTABLES = [
        ("script list", parseScriptList),
        ("feature list", parseFeatureList),
        ("lookup list", parseLookupList),
    ]
    offsets = []
    for description, parser in SUBTABLES:
        name = description.title().replace(" ", "")
        offset = UInt16(
            self, name[0].lower() + name[1:], "Offset to %s table" % description
        )
        yield offset
        offsets.append((offset.value, parser))
    if self["min_ver"].value == 1:
        offset = UInt32(
            self, "featureVariationsOffset", "Offset to feature variations table"
        )
        offsets.append((offset.value, parseFeatureVariationsTable))

    offsets.sort(key=lambda field: field[0])
    padding = self.seekByte(offsets[0][0], null=True)
    if padding:
        yield padding
    lastOffset, first_parser = offsets[0]
    for offset, parser in offsets[1:]:
        # yield parser(self)
        yield RawBytes(self, "content", offset - lastOffset)
        lastOffset = offset


class Table(FieldSet):
    TAG_INFO = {
        "DSIG": ("DSIG", "Digital Signature", parseDSIG),
        "GSUB": ("GSUB", "Glyph Substitutions", parseGSUB),
        "avar": ("avar", "Axis variation table", parseAvar),
        "cmap": ("cmap", "Character to Glyph Index Mapping", parseCmap),
        "fvar": ("fvar", "Font variations table", parseFvar),
        "head": ("header", "Font header", parseFontHeader),
        "hhea": ("hhea", "Horizontal Header", parseHhea),
        "maxp": ("maxp", "Maximum Profile", parseMaxp),
        "name": ("names", "Names", parseNames),
        "OS/2": ("OS_2", "OS/2 and Windows Metrics", parseOS2),
        "post": ("post", "PostScript", parsePost),
    }

    def __init__(self, parent, name, table, **kw):
        FieldSet.__init__(self, parent, name, **kw)
        self.table = table
        tag = table["tag"].value
        if tag in self.TAG_INFO:
            self._name, self._description, self.parser = self.TAG_INFO[tag]
        else:
            self.parser = None

    def createFields(self):
        if self.parser:
            yield from self.parser(self)
        else:
            yield RawBytes(self, "content", self.size // 8)

    def createDescription(self):
        return "Table %s (%s)" % (self.table["tag"].value, self.table.path)


class TrueTypeFontFile(Parser):
    endian = BIG_ENDIAN
    PARSER_TAGS = {
        "id": "ttf",
        "category": "misc",
        "file_ext": ("ttf",),
        "min_size": 10 * 8,  # FIXME
        "description": "TrueType font",
    }

    def validate(self):
        if self["maj_ver"].value == 1 and self["min_ver"].value == 0:
            pass
        elif self["maj_ver"].value == 0x4F54 and self["min_ver"].value == 0x544F:
            pass
        else:
            return "Invalid version (%u.%u)" % (
                self["maj_ver"].value,
                self["min_ver"].value,
            )
        if not (MIN_NB_TABLE <= self["nb_table"].value <= MAX_NB_TABLE):
            return "Invalid number of table (%u)" % self["nb_table"].value
        return True

    def createFields(self):
        yield UInt16(self, "maj_ver", "Major version")
        yield UInt16(self, "min_ver", "Minor version")
        yield UInt16(self, "nb_table")
        yield UInt16(self, "search_range")
        yield UInt16(self, "entry_selector")
        yield UInt16(self, "range_shift")
        tables = []
        for index in range(self["nb_table"].value):
            table = TableHeader(self, "table_hdr[]")
            yield table
            tables.append(table)
        tables.sort(key=lambda field: field["offset"].value)
        for table in tables:
            padding = self.seekByte(table["offset"].value, null=True)
            if padding:
                yield padding
            size = table["size"].value
            if size:
                yield Table(self, "table[]", table, size=size * 8)
        padding = self.seekBit(self.size, null=True)
        if padding:
            yield padding



"""
ISO 9660 (cdrom) file system parser.

Documents:
- Standard ECMA-119 (december 1987)
  http://www.nondot.org/sabre/os/files/FileSystems/iso9660.pdf

Author: Victor Stinner
Creation: 11 july 2006
"""

from hachoir.parser import Parser
from hachoir.field import (FieldSet, ParserError,
                           UInt8, UInt32, UInt64, Enum,
                           NullBytes, RawBytes, String)
from hachoir.core.endian import LITTLE_ENDIAN, BIG_ENDIAN


class PrimaryVolumeDescriptor(FieldSet):
    static_size = 2041 * 8

    def createFields(self):
        yield NullBytes(self, "unused[]", 1)
        yield String(self, "system_id", 32, "System identifier", strip=" ")
        yield String(self, "volume_id", 32, "Volume identifier", strip=" ")
        yield NullBytes(self, "unused[]", 8)
        yield UInt64(self, "space_size", "Volume space size")
        yield NullBytes(self, "unused[]", 32)
        yield UInt32(self, "set_size", "Volume set size")
        yield UInt32(self, "seq_num", "Sequence number")
        yield UInt32(self, "block_size", "Block size")
        yield UInt64(self, "path_table_size", "Path table size")
        yield UInt32(self, "occu_lpath", "Location of Occurrence of Type L Path Table")
        yield UInt32(self, "opt_lpath", "Location of Optional of Type L Path Table")
        yield UInt32(self, "occu_mpath", "Location of Occurrence of Type M Path Table")
        yield UInt32(self, "opt_mpath", "Location of Optional of Type M Path Table")
        yield RawBytes(self, "root", 34, "Directory Record for Root Directory")
        yield String(self, "vol_set_id", 128, "Volume set identifier", strip=" ")
        yield String(self, "publisher", 128, "Publisher identifier", strip=" ")
        yield String(self, "data_preparer", 128, "Data preparer identifier", strip=" ")
        yield String(self, "application", 128, "Application identifier", strip=" ")
        yield String(self, "copyright", 37, "Copyright file identifier", strip=" ")
        yield String(self, "abstract", 37, "Abstract file identifier", strip=" ")
        yield String(self, "biographic", 37, "Biographic file identifier", strip=" ")
        yield String(self, "creation_ts", 17, "Creation date and time", strip=" ")
        yield String(self, "modification_ts", 17, "Modification date and time", strip=" ")
        yield String(self, "expiration_ts", 17, "Expiration date and time", strip=" ")
        yield String(self, "effective_ts", 17, "Effective date and time", strip=" ")
        yield UInt8(self, "struct_ver", "Structure version")
        yield NullBytes(self, "unused[]", 1)
        yield String(self, "app_use", 512, "Application use", strip=" \0")
        yield NullBytes(self, "unused[]", 653)


class BootRecord(FieldSet):
    static_size = 2041 * 8

    def createFields(self):
        yield String(self, "sys_id", 31, "Boot system identifier", strip="\0")
        yield String(self, "boot_id", 31, "Boot identifier", strip="\0")
        yield RawBytes(self, "system_use", 1979, "Boot system use")


class Terminator(FieldSet):
    static_size = 2041 * 8

    def createFields(self):
        yield NullBytes(self, "null", 2041)


class Volume(FieldSet):
    endian = BIG_ENDIAN
    TERMINATOR = 255
    type_name = {
        0: "Boot Record",
        1: "Primary Volume Descriptor",
        2: "Supplementary Volume Descriptor",
        3: "Volume Partition Descriptor",
        TERMINATOR: "Volume Descriptor Set Terminator",
    }
    static_size = 2048 * 8
    content_handler = {
        0: BootRecord,
        1: PrimaryVolumeDescriptor,
        TERMINATOR: Terminator,
    }

    def createFields(self):
        yield Enum(UInt8(self, "type", "Volume descriptor type"), self.type_name)
        yield RawBytes(self, "signature", 5, "ISO 9960 signature (CD001)")
        if self["signature"].value != b"CD001":
            raise ParserError("Invalid ISO 9960 volume signature")
        yield UInt8(self, "version", "Volume descriptor version")
        cls = self.content_handler.get(self["type"].value, None)
        if cls:
            yield cls(self, "content")
        else:
            yield RawBytes(self, "raw_content", 2048 - 7)


class ISO9660(Parser):
    endian = LITTLE_ENDIAN
    MAGIC = b"\x01CD001"
    NULL_BYTES = 0x8000
    PARSER_TAGS = {
        "id": "iso9660",
        "category": "file_system",
        "description": "ISO 9660 file system",
        "min_size": (NULL_BYTES + 6) * 8,
        "magic": ((MAGIC, NULL_BYTES * 8),),
    }

    def validate(self):
        if self.stream.readBytes(self.NULL_BYTES * 8, len(self.MAGIC)) != self.MAGIC:
            return "Invalid signature"
        return True

    def createFields(self):
        yield self.seekByte(self.NULL_BYTES, null=True)

        while True:
            volume = Volume(self, "volume[]")
            yield volume
            if volume["type"].value == Volume.TERMINATOR:
                break

        if self.current_size < self._size:
            yield self.seekBit(self._size, "end")



"""
Microsoft Archive parser

Author: Victor Stinner
Creation date: 2007-03-04
"""

from hachoir.parser import Parser
from hachoir.field import FieldSet, String, UInt32, SubFile
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.core.text_handler import textHandler, filesizeHandler, hexadecimal

MAX_NB_FILE = 100000


class FileIndex(FieldSet):
    static_size = 68 * 8

    def createFields(self):
        yield String(self, "filename", 56, truncate="\0", charset="ASCII")
        yield filesizeHandler(UInt32(self, "filesize"))
        yield textHandler(UInt32(self, "crc32"), hexadecimal)
        yield UInt32(self, "offset")

    def createDescription(self):
        return "File %s (%s) at %s" % (
            self["filename"].value, self["filesize"].display, self["offset"].value)


class MarFile(Parser):
    MAGIC = b"MARC"
    PARSER_TAGS = {
        "id": "mar",
        "category": "archive",
        "file_ext": ("mar",),
        "min_size": 80 * 8,  # At least one file index
        "magic": ((MAGIC, 0),),
        "description": "Microsoft Archive",
    }
    endian = LITTLE_ENDIAN

    def validate(self):
        if self.stream.readBytes(0, 4) != self.MAGIC:
            return "Invalid magic"
        if self["version"].value != 3:
            return "Invalid version"
        if not (1 <= self["nb_file"].value <= MAX_NB_FILE):
            return "Invalid number of file"
        return True

    def createFields(self):
        yield String(self, "magic", 4, "File signature (MARC)", charset="ASCII")
        yield UInt32(self, "version")
        yield UInt32(self, "nb_file")
        files = []
        for index in range(self["nb_file"].value):
            item = FileIndex(self, "file[]")
            yield item
            if item["filesize"].value:
                files.append(item)
        files.sort(key=lambda item: item["offset"].value)
        for index in files:
            padding = self.seekByte(index["offset"].value)
            if padding:
                yield padding
            size = index["filesize"].value
            desc = "File %s" % index["filename"].value
            yield SubFile(self, "data[]", size, desc, filename=index["filename"].value)



import gc

# ---- Default implementation when resource is missing ----------------------
PAGE_SIZE = 4096


def getMemoryLimit():
    """
    Get current memory limit in bytes.

    Return None on error.
    """
    return None


def setMemoryLimit(max_mem):
    """
    Set memory limit in bytes.
    Use value 'None' to disable memory limit.

    Return True if limit is set, False on error.
    """
    return False


def getMemorySize():
    """
    Read currenet process memory size: size of available virtual memory.
    This value is NOT the real memory usage.

    This function only works on Linux (use /proc/self/statm file).
    """
    try:
        statm = open('/proc/self/statm').readline().split()
    except IOError:
        return None
    return int(statm[0]) * PAGE_SIZE


def clearCaches():
    """
    Try to clear all caches: call gc.collect() (Python garbage collector).
    """
    gc.collect()
    # import re; re.purge()


try:
    # ---- 'resource' implementation -----------------------------------
    from resource import getpagesize, getrlimit, setrlimit, RLIMIT_AS

    PAGE_SIZE = getpagesize()

    def getMemoryLimit():  # noqa
        try:
            limit = getrlimit(RLIMIT_AS)[0]
            if 0 < limit:
                limit *= PAGE_SIZE
            return limit
        except ValueError:
            return None

    def setMemoryLimit(max_mem):  # noqa
        if max_mem is None:
            max_mem = -1
        try:
            setrlimit(RLIMIT_AS, (max_mem, -1))
            return True
        except ValueError:
            return False
except ImportError:
    pass


def limitedMemory(limit, func, *args, **kw):
    """
    Limit memory grow when calling func(*args, **kw):
    restrict memory grow to 'limit' bytes.

    Use try/except MemoryError to catch the error.
    """
    # First step: clear cache to gain memory
    clearCaches()

    # Get total program size
    max_rss = getMemorySize()
    if max_rss is not None:
        # Get old limit and then set our new memory limit
        old_limit = getMemoryLimit()
        limit = max_rss + limit
        limited = setMemoryLimit(limit)
    else:
        limited = False

    try:
        # Call function
        return func(*args, **kw)
    finally:
        # and unset our memory limit
        if limited:
            setMemoryLimit(old_limit)

        # After calling the function: clear all caches
        clearCaches()



from hachoir.core.i18n import getTerminalCharset
from hachoir.core.cmd_line import getHachoirOptions, configureHachoir
from hachoir.field import isString
from hachoir.core.benchmark import Benchmark
from hachoir.core.error import error
from hachoir.stream import InputStreamError
from hachoir.core.tools import makePrintable
from hachoir.parser import createParser
from hachoir.core.cmd_line import displayVersion
from optparse import OptionGroup, OptionParser
import errno
import sys


def countChildren(field_set):
    count = 0
    for field in field_set.fields.values:
        if field.is_field_set:
            count += countChildren(field)
        count += 1
    return count


def displayParserStat(parser):
    print("Parser: %s children"
          % countChildren(parser))


def parseOptions():
    parser = OptionParser(usage="%prog [options] "
                                "pattern filename [filename2 ...]")

    common = OptionGroup(parser, "Grep", "Option of grep")
    common.add_option("--percent", help="Display percent",
                      action="store_true", default=False)
    common.add_option("--no-addr", help="Don't display address",
                      action="store_true", default=False)
    common.add_option("--no-value", help="Don't display value",
                      action="store_true", default=False)
    common.add_option("--case", help="Search is case sensitive",
                      action="store_true", default=False)
    common.add_option("--path", help="Display path",
                      action="store_true", default=False)
    common.add_option("--all", help="Match all (just extract strings)",
                      action="store_true", default=False)
    common.add_option("--bench", help="Run benchmark",
                      action="store_true", default=False)
    common.add_option("--version", help="Display version and exit",
                      action="callback", callback=displayVersion)
    parser.add_option_group(common)

    hachoir = getHachoirOptions(parser)
    parser.add_option_group(hachoir)

    values, arguments = parser.parse_args()
    if values.all or values.bench:
        if len(arguments) < 1:
            parser.print_help()
            sys.exit(1)
        pattern = None
        filenames = arguments
    else:
        if len(arguments) < 2:
            parser.print_help()
            sys.exit(1)
        pattern = arguments[0]
        filenames = arguments[1:]
    return values, pattern, filenames


class Grep:

    def __init__(self):
        self.pattern = None
        self.case_sensitive = True

    def grep(self, fieldset):
        for field in fieldset:
            if field.is_field_set:
                self.grep(field)
                field.reset()
            elif isString(field) and self.match(field):
                self.onMatch(field)

    def match(self, field):
        value = field.value
        if len(value) == 0:
            return False
        if not self.pattern:
            return True
        if not self.case_sensitive:
            value = value.lower()
        return self.pattern in value

    def onMatch(self, field):
        raise NotImplementedError()


class ConsoleGrep(Grep):

    def __init__(self):
        Grep.__init__(self)
        self.term_charset = getTerminalCharset()
        self.display = True
        self.display_filename = True
        self.display_address = True
        self.display_value = True
        self.display_path = False
        self.display_percent = False
        self.filename = None

    def onMatch(self, field):
        if not self.display:
            return
        text = []
        if self.display_percent or self.display_address:
            addr = field.absolute_address
        if self.display_filename:
            filename = makePrintable(self.filename, self.term_charset)
            text.append(filename)
        if self.display_address:
            if (addr % 8) == 0:
                text.append(str(addr // 8))
            else:
                text.append("%u.%u" % (addr // 8, addr % 8))
        if self.display_path:
            text.append(field.path)
        if self.display_value:
            value = field.value
            value = makePrintable(value, self.term_charset)
            text.append(value)
        if not text:
            return
        text = ":".join(text)
        if self.display_percent:
            percent = float(addr) * 100 / field.parent.root.size
            sys.stdout.flush()
            sys.stderr.write("[%02.1f%%] " % percent)
            sys.stderr.flush()
        print(text)

    def searchFile(self, filename, pattern, case_sensitive=True):
        self.filename = filename
        self.case_sensitive = case_sensitive
        if pattern and not self.case_sensitive:
            pattern = pattern.lower()
        self.pattern = pattern

        try:
            self.parser = createParser(self.filename)
        except InputStreamError as err:
            error("Unable to open file: %s" % err)
            sys.exit(1)
        if not self.parser:
            error("Unable to parse file: %s" % self.filename)
            sys.exit(1)

        with self.parser:
            try:
                self.grep(self.parser)
            except IOError as err:
                if err[0] == errno.EPIPE:
                    # Ignore broken PIPE error
                    return
                else:
                    raise


def runGrep(values, pattern, filenames):
    grep = ConsoleGrep()
    grep.display_filename = (1 < len(filenames))
    grep.display_address = not values.no_addr
    grep.display_path = values.path
    grep.display_value = not values.no_value
    grep.display_percent = values.percent
    grep.display = not values.bench
    for filename in filenames:
        grep.searchFile(filename, pattern, case_sensitive=values.case)


def main():
    try:
        values, pattern, filenames = parseOptions()
        configureHachoir(values)
        if values.bench:
            bench = Benchmark()
            bench.run(runGrep, values, pattern, filenames)
        else:
            runGrep(values, pattern, filenames)
    except KeyboardInterrupt:
        print("Program interrupted (CTRL+C).")
        sys.exit(1)


if __name__ == '__main__':
    main()



"""
The ScreamTracker 3.0x module format description for .s3m files.

Documents:
- Search s3m on Wotsit
  http://www.wotsit.org/

Author: Christophe GISQUET <christophe.gisquet@free.fr>
Creation: 11th February 2007
"""

from hachoir.parser import Parser
from hachoir.field import (StaticFieldSet, FieldSet, Field,
                           Bit, Bits,
                           UInt32, UInt16, UInt8, Enum,
                           PaddingBytes, RawBytes, NullBytes,
                           String, GenericVector, ParserError)
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.core.text_handler import textHandler, hexadecimal
from hachoir.core.tools import alignValue


class Chunk:

    def __init__(self, cls, name, offset, size, *args):
        # Todo: swap and have None=unknown instead of now: 0=unknown
        assert size is not None and size >= 0
        self.cls = cls
        self.name = name
        self.offset = offset
        self.size = size
        self.args = args


class ChunkIndexer:

    def __init__(self):
        self.chunks = []

    # Check if a chunk fits
    def canHouse(self, chunk, index):
        if index > 1:
            if chunk.offset + chunk.size > self.chunks[index - 1].offset:
                return False
        # We could test now that it fits in the memory
        return True

    # Farthest element is last
    def addChunk(self, new_chunk):
        index = 0
        # Find first chunk whose value is bigger
        while index < len(self.chunks):
            offset = self.chunks[index].offset
            if offset < new_chunk.offset:
                if not self.canHouse(new_chunk, index):
                    raise ParserError("Chunk '%s' doesn't fit!" %
                                      new_chunk.name)
                self.chunks.insert(index, new_chunk)
                return
            index += 1

        # Not found or empty
        # We could at least check that it fits in the memory
        self.chunks.append(new_chunk)

    def yieldChunks(self, obj):
        while len(self.chunks) > 0:
            chunk = self.chunks.pop()
            current_pos = obj.current_size // 8

            # Check if padding needed
            size = chunk.offset - current_pos
            if size > 0:
                obj.info("Padding of %u bytes needed: curr=%u offset=%u" %
                         (size, current_pos, chunk.offset))
                yield PaddingBytes(obj, "padding[]", size)
                current_pos = obj.current_size // 8

            # Find resynch point if needed
            count = 0
            old_off = chunk.offset
            while chunk.offset < current_pos:
                count += 1
                chunk = self.chunks.pop()
                # Unfortunaly, we also pass the underlying chunks
                if chunk is None:
                    obj.info("Couldn't resynch: %u object skipped to reach %u" %
                             (count, current_pos))
                    return

            # Resynch
            size = chunk.offset - current_pos
            if size > 0:
                obj.info("Skipped %u objects to resynch to %u; chunk offset: %u->%u" %
                         (count, current_pos, old_off, chunk.offset))
                yield RawBytes(obj, "resynch[]", size)

            # Yield
            obj.info("Yielding element of size %u at offset %u" %
                     (chunk.size, chunk.offset))
            field = chunk.cls(obj, chunk.name, chunk.size, *chunk.args)
            yield field

            if hasattr(field, "getSubChunks"):
                for sub_chunk in field.getSubChunks():
                    obj.info("Adding sub chunk: position=%u size=%u name='%s'" %
                             (sub_chunk.offset, sub_chunk.size, sub_chunk.name))
                    self.addChunk(sub_chunk)

            # Let missing padding be done by next chunk


class S3MFlags(StaticFieldSet):
    format = (
        (Bit, "st2_vibrato", "Vibrato (File version 1/ScreamTrack 2)"),
        (Bit, "st2_tempo", "Tempo (File version 1/ScreamTrack 2)"),
        (Bit, "amiga_slides", "Amiga slides (File version 1/ScreamTrack 2)"),
        (Bit, "zero_vol_opt",
         "Automatically turn off looping notes whose volume is zero for >2 note rows"),
        (Bit, "amiga_limits", "Disallow notes beyond Amiga hardware specs"),
        (Bit, "sb_processing", "Enable filter/SFX with SoundBlaster"),
        (Bit, "vol_slide", "Volume slide also performed on first row"),
        (Bit, "extended", "Special custom data in file"),
        (Bits, "unused[]", 8)
    )


def parseChannelType(val):
    val = val.value
    if val < 8:
        return "Left Sample Channel %u" % val
    if val < 16:
        return "Right Sample Channel %u" % (val - 8)
    if val < 32:
        return "Adlib channel %u" % (val - 16)
    return "Value %u unknown" % val


class ChannelSettings(FieldSet):
    static_size = 8

    def createFields(self):
        yield textHandler(Bits(self, "type", 7), parseChannelType)
        yield Bit(self, "enabled")


class ChannelPanning(FieldSet):
    static_size = 8

    def createFields(self):
        yield Bits(self, "default_position", 4, "Default pan position")
        yield Bit(self, "reserved[]")
        yield Bit(self, "use_default", "Bits 0:3 specify default position")
        yield Bits(self, "reserved[]", 2)

# Provide an automatic constructor


class SizeFieldSet(FieldSet):
    """
    Provide an automatic constructor for a sized field that can be aligned
    on byte positions according to ALIGN.

    Size is ignored if static_size is set. Real size is stored
    for convenience, but beware, it is not in bits, but in bytes.

    Field can be automatically padded, unless:
    - size is 0 (unknown, so padding doesn't make sense)
    - it shouldn't be aligned

    If it shouldn't be aligned, two solutions:
    - change _size to another value than the one found through aligment.
    - derive a class with ALIGN = 0.
    """
    ALIGN = 16

    def __init__(self, parent, name, size, desc=None):
        FieldSet.__init__(self, parent, name, desc)
        if size:
            self.real_size = size
            if self.static_size is None:
                self.setCheckedSizes(size)

    def setCheckedSizes(self, size):
        # First set size so that end is aligned, if needed
        self.real_size = size
        size *= 8
        if self.ALIGN:
            size = alignValue(self.absolute_address + size, 8 * self.ALIGN) \
                - self.absolute_address

        if self._parent._size:
            if self._parent.current_size + size > self._parent._size:
                size = self._parent._size - self._parent.current_size

        self._size = size

    def createFields(self):
        yield from self.createUnpaddedFields()
        size = (self._size - self.current_size) // 8
        if size > 0:
            yield PaddingBytes(self, "padding", size)


class Header(SizeFieldSet):

    def createDescription(self):
        return "%s (%u patterns, %u instruments)" % \
               (self["title"].value, self["num_patterns"].value,
                self["num_instruments"].value)

    def createValue(self):
        return self["title"].value

    # Header fields may have to be padded - specify static_size
    # or modify _size in a derived class if never.
    def createUnpaddedFields(self):
        yield String(self, "title", 28, strip='\0')
        yield textHandler(UInt8(self, "marker[]"), hexadecimal)
        yield from self.getFileVersionField()

        yield UInt16(self, "num_orders")
        yield UInt16(self, "num_instruments")
        yield UInt16(self, "num_patterns")

        yield from self.getFirstProperties()

        yield String(self, "marker[]", 4)
        yield from self.getLastProperties()

        yield GenericVector(self, "channel_settings", 32,
                            ChannelSettings, "channel")

        # Orders
        yield GenericVector(self, "orders", self.getNumOrders(), UInt8, "order")

        yield from self.getHeaderEndFields()


class S3MHeader(Header):
    """
          0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F
        +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
  0000: | Song name, max 28 chars (end with NUL (0))                    |
        +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
  0010: |                                               |1Ah|Typ| x | x |
        +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
  0020: |OrdNum |InsNum |PatNum | Flags | Cwt/v | Ffi   |'S'|'C'|'R'|'M'|
        +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
  0030: |g.v|i.s|i.t|m.v|u.c|d.p| x | x | x | x | x | x | x | x |Special|
        +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
  0040: |Channel settings for 32 channels, 255=unused,+128=disabled     |
        +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
  0050: |                                                               |
        +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
  0060: |Orders; length=OrdNum (should be even)                         |
        +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
  xxx1: |Parapointers to instruments; length=InsNum*2                   |
        +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
  xxx2: |Parapointers to patterns; length=PatNum*2                      |
        +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
  xxx3: |Channel default pan positions                                  |
        +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
        xxx1=70h+orders
        xxx2=70h+orders+instruments*2
        xxx3=70h+orders+instruments*2+patterns*2
    """

    def __init__(self, parent, name, size, desc=None):
        Header.__init__(self, parent, name, size, desc)

        # Overwrite real_size
        size = 0x60 + self["num_orders"].value + \
            2 * (self["num_instruments"].value + self["num_patterns"].value)
        if self["panning_info"].value == 252:
            size += 32

        # Deduce size for SizeFieldSet
        self.setCheckedSizes(size)

    def getFileVersionField(self):
        yield UInt8(self, "type")
        yield RawBytes(self, "reserved[]", 2)

    def getFirstProperties(self):
        yield S3MFlags(self, "flags")
        yield UInt8(self, "creation_version_minor")
        yield Bits(self, "creation_version_major", 4)
        yield Bits(self, "creation_version_unknown", 4, "(=1)")
        yield UInt16(self, "format_version")

    def getLastProperties(self):
        yield UInt8(self, "glob_vol", "Global volume")
        yield UInt8(self, "init_speed", "Initial speed (command A)")
        yield UInt8(self, "init_tempo", "Initial tempo (command T)")
        yield Bits(self, "volume", 7)
        yield Bit(self, "stereo")
        yield UInt8(self, "click_removal", "Number of GUS channels to run to prevent clicks")
        yield UInt8(self, "panning_info")
        yield RawBytes(self, "reserved[]", 8)
        yield UInt16(self, "custom_data_parapointer",
                     "Parapointer to special custom data (not used by ST3.01)")

    def getNumOrders(self):
        return self["num_orders"].value

    def getHeaderEndFields(self):
        instr = self["num_instruments"].value
        patterns = self["num_patterns"].value
        # File pointers
        if instr > 0:
            yield GenericVector(self, "instr_pptr", instr, UInt16, "offset")
        if patterns > 0:
            yield GenericVector(self, "pattern_pptr", patterns, UInt16, "offset")

        # S3M 3.20 extension
        if self["creation_version_major"].value >= 3 \
                and self["creation_version_minor"].value >= 0x20 \
                and self["panning_info"].value == 252:
            yield GenericVector(self, "channel_panning", 32, ChannelPanning, "channel")

        # Padding required for 16B alignment
        size = self._size - self.current_size
        if size > 0:
            yield PaddingBytes(self, "padding", size // 8)

    def getSubChunks(self):
        # Instruments -  no warranty that they are concatenated
        for index in range(self["num_instruments"].value):
            yield Chunk(S3MInstrument, "instrument[]",
                        16 * self["instr_pptr/offset[%u]" % index].value,
                        S3MInstrument.static_size // 8)

        # Patterns - size unknown but listed in their headers
        for index in range(self["num_patterns"].value):
            yield Chunk(S3MPattern, "pattern[]",
                        16 * self["pattern_pptr/offset[%u]" % index].value, 0)


class PTMHeader(Header):
    # static_size should prime over _size, right?
    static_size = 8 * 608

    def getTrackerVersion(self, val):
        val = val.value
        return "ProTracker x%04X" % val

    def getFileVersionField(self):
        yield UInt16(self, "type")
        yield RawBytes(self, "reserved[]", 1)

    def getFirstProperties(self):
        yield UInt16(self, "channels")
        yield UInt16(self, "flags")  # 0 => NullBytes
        yield UInt16(self, "reserved[]")

    def getLastProperties(self):
        yield RawBytes(self, "reserved[]", 16)

    def getNumOrders(self):
        return 256

    def getHeaderEndFields(self):
        yield GenericVector(self, "pattern_pptr", 128, UInt16, "offset")

    def getSubChunks(self):
        # It goes like this in the BS: patterns->instruments->instr. samples

        if self._parent._size:
            min_off = self.absolute_address + self._parent._size
        else:
            min_off = 99999999999

        # Instruments and minimal end position for last pattern
        count = self["num_instruments"].value
        addr = self.absolute_address
        for index in range(count):
            offset = (self.static_size + index *
                      PTMInstrument.static_size) // 8
            yield Chunk(PTMInstrument, "instrument[]", offset,
                        PTMInstrument.static_size // 8)
            offset = self.stream.readBits(
                addr + 8 * (offset + 18), 32, LITTLE_ENDIAN)
            min_off = min(min_off, offset)

        # Patterns
        count = self["num_patterns"].value
        prev_off = 16 * self["pattern_pptr/offset[0]"].value
        for index in range(1, count):
            offset = 16 * self["pattern_pptr/offset[%u]" % index].value
            yield Chunk(PTMPattern, "pattern[]", prev_off, offset - prev_off)
            prev_off = offset

        # Difficult to account for
        yield Chunk(PTMPattern, "pattern[]", prev_off, min_off - prev_off)


class SampleFlags(StaticFieldSet):
    format = (
        (Bit, "loop_on"),
        (Bit, "stereo", "Sample size will be 2*length"),
        (Bit, "16bits", "16b sample, Intel LO-HI byteorder"),
        (Bits, "unused", 5)
    )


class S3MUInt24(Field):
    static_size = 24

    def __init__(self, parent, name, desc=None):
        Field.__init__(self, parent, name, size=24, description=desc)
        addr = self.absolute_address
        val = parent.stream.readBits(addr, 8, LITTLE_ENDIAN) << 20
        val += parent.stream.readBits(addr + 8, 16, LITTLE_ENDIAN) << 4
        self.createValue = lambda: val


class SampleData(SizeFieldSet):

    def createUnpaddedFields(self):
        yield RawBytes(self, "data", self.real_size)


class PTMSampleData(SampleData):
    ALIGN = 0


class Instrument(SizeFieldSet):
    static_size = 8 * 0x50

    def createDescription(self):
        info = [self["c4_speed"].display]
        if "flags/stereo" in self:
            if self["flags/stereo"].value:
                info.append("stereo")
            else:
                info.append("mono")
        info.append("%u bits" % self.getSampleBits())
        return ", ".join(info)

    # Structure knows its size and doesn't need padding anyway, so
    # overwrite base member: no need to go through it.
    def createFields(self):
        yield self.getType()
        yield String(self, "filename", 12, strip='\0')

        yield from self.getInstrumentFields()

        yield String(self, "name", 28, strip='\0')
        yield String(self, "marker", 4, "Either 'SCRS' or '(empty)'", strip='\0')

    def createValue(self):
        return self["name"].value


class S3MInstrument(Instrument):
    """
    In fact a sample. Description follows:

          0   1   2   3   4   5   6   7   8   9   A   B   C   D   E   F
        +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
  0000: |[T]| Dos filename (12345678.ABC)                   |    MemSeg |
        +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
  0010: |Length |HI:leng|LoopBeg|HI:LBeg|LoopEnd|HI:Lend|Vol| x |[P]|[F]|
        +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
  0020: |C2Spd  |HI:C2sp| x | x | x | x |Int:Gp |Int:512|Int:lastused   |
        +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
  0030: | Sample name, 28 characters max... (incl. NUL)                 |
        +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
  0040: | ...sample name...                             |'S'|'C'|'R'|'S'|
        +---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+---+
  xxxx: sampledata
    """
    MAGIC = b"SCRS"
    PACKING = {0: "Unpacked", 1: "DP30ADPCM"}
    TYPE = {0: "Unknown", 1: "Sample", 2: "adlib melody", 3: "adlib drum2"}

    def getType(self):
        return Enum(UInt8(self, "type"), self.TYPE)

    def getSampleBits(self):
        return 8 * (1 + self["flags/16bits"].value)

    def getInstrumentFields(self):
        yield S3MUInt24(self, "sample_offset")
        yield UInt32(self, "sample_size")
        yield UInt32(self, "loop_begin")
        yield UInt32(self, "loop_end")
        yield UInt8(self, "volume")
        yield UInt8(self, "reserved[]")
        yield Enum(UInt8(self, "packing"), self.PACKING)
        yield SampleFlags(self, "flags")
        yield UInt32(self, "c4_speed", "Frequency for middle C note")
        yield UInt32(self, "reserved[]", 4)
        yield UInt16(self, "internal[]", "Sample address in GUS memory")
        yield UInt16(self, "internal[]", "Flags for SoundBlaster loop expansion")
        yield UInt32(self, "internal[]", "Last used position (SB)")

    def getSubChunks(self):
        size = self["sample_size"].value
        if self["flags/stereo"].value:
            size *= 2
        if self["flags/16bits"].value:
            size *= 2
        yield Chunk(SampleData, "sample_data[]",
                    self["sample_offset"].value, size)


class PTMType(FieldSet):
    TYPES = {0: "No sample", 1: "Regular",
             2: "OPL2/OPL2 instrument", 3: "MIDI instrument"}
    static_size = 8

    def createFields(self):
        yield Bits(self, "unused", 2)
        yield Bit(self, "is_tonable")
        yield Bit(self, "16bits")
        yield Bit(self, "loop_bidir")
        yield Bit(self, "loop")
        yield Enum(Bits(self, "origin", 2), self.TYPES)

#  class PTMType(StaticFieldSet):
#  format = (
# #        (Bits, "unused", 2),
# #        (Bit, "is_tonable"),
# #        (Bit, "16bits"),
# #        (Bit, "loop_bidir"),
# #        (Bit, "loop"),
# #        (Bits, "origin", 2),
#  )


class PTMInstrument(Instrument):
    MAGIC = b"PTMI"
    ALIGN = 0

    def getType(self):
        return PTMType(self, "flags")  # Hack to have more common code

    # PTM doesn't pretend to manage 16bits
    def getSampleBits(self):
        return 8

    def getInstrumentFields(self):
        yield UInt8(self, "volume")
        yield UInt16(self, "c4_speed")
        yield UInt16(self, "sample_segment")
        yield UInt32(self, "sample_offset")
        yield UInt32(self, "sample_size")
        yield UInt32(self, "loop_begin")
        yield UInt32(self, "loop_end")
        yield UInt32(self, "gus_begin")
        yield UInt32(self, "gus_loop_start")
        yield UInt32(self, "gus_loop_end")
        yield textHandler(UInt8(self, "gus_loop_flags"), hexadecimal)
        yield UInt8(self, "reserved[]")  # Should be 0

    def getSubChunks(self):
        # Samples are NOT padded, and the size is already the correct one
        size = self["sample_size"].value
        if size:
            yield Chunk(PTMSampleData, "sample_data[]", self["sample_offset"].value, size)


class S3MNoteInfo(StaticFieldSet):
    """
0=end of row
&31=channel
&32=follows;  BYTE:note, BYTE:instrument
&64=follows;  BYTE:volume
&128=follows; BYTE:command, BYTE:info
    """
    format = (
        (Bits, "channel", 5),
        (Bit, "has_note"),
        (Bit, "has_volume"),
        (Bit, "has_effect")
    )


class PTMNoteInfo(StaticFieldSet):
    format = (
        (Bits, "channel", 5),
        (Bit, "has_note"),
        (Bit, "has_effect"),
        (Bit, "has_volume")
    )


class Note(FieldSet):

    def createFields(self):
        # Used by Row to check if end of Row
        info = self.NOTE_INFO(self, "info")
        yield info
        if info["has_note"].value:
            yield UInt8(self, "note")
            yield UInt8(self, "instrument")
        if info["has_volume"].value:
            yield UInt8(self, "volume")
        if info["has_effect"].value:
            yield UInt8(self, "effect")
            yield UInt8(self, "param")


class S3MNote(Note):
    NOTE_INFO = S3MNoteInfo


class PTMNote(Note):
    NOTE_INFO = PTMNoteInfo


class Row(FieldSet):

    def createFields(self):
        addr = self.absolute_address
        while True:
            # Check empty note
            byte = self.stream.readBits(addr, 8, self.endian)
            if not byte:
                yield NullBytes(self, "terminator", 1)
                return

            note = self.NOTE(self, "note[]")
            yield note
            addr += note.size


class S3MRow(Row):
    NOTE = S3MNote


class PTMRow(Row):
    NOTE = PTMNote


class Pattern(SizeFieldSet):

    def createUnpaddedFields(self):
        count = 0
        while count < 64 and not self.eof:
            yield self.ROW(self, "row[]")
            count += 1


class S3MPattern(Pattern):
    ROW = S3MRow

    def __init__(self, parent, name, size, desc=None):
        Pattern.__init__(self, parent, name, size, desc)

        # Get real_size from header
        addr = self.absolute_address
        size = self.stream.readBits(addr, 16, LITTLE_ENDIAN)
        self.setCheckedSizes(size)


class PTMPattern(Pattern):
    ROW = PTMRow


class Module(Parser):
    # MARKER / HEADER are defined in derived classes
    endian = LITTLE_ENDIAN

    def validate(self):
        marker = self.stream.readBits(0x1C * 8, 8, LITTLE_ENDIAN)
        if marker != 0x1A:
            return "Invalid start marker %u" % marker
        marker = self.stream.readBytes(0x2C * 8, 4)
        if marker != self.MARKER:
            return "Invalid marker %s!=%s" % (marker, self.MARKER)
        return True

    def createFields(self):
        # Index chunks
        indexer = ChunkIndexer()
        # Add header - at least 0x50 bytes
        indexer.addChunk(Chunk(self.HEADER, "header", 0, 0x50))
        yield from indexer.yieldChunks(self)


class S3MModule(Module):
    PARSER_TAGS = {
        "id": "s3m",
        "category": "audio",
        "file_ext": ("s3m",),
        "mime": ('audio/s3m', 'audio/x-s3m'),
        "min_size": 64 * 8,
        "description": "ScreamTracker3 module"
    }
    MARKER = b"SCRM"
    HEADER = S3MHeader

#  def createContentSize(self):
# #        hdr = Header(self, "header")
# #        max_offset = hdr._size//8

# #        instr_size = Instrument._size//8
#  for index in xrange(self["header/num_instruments"].value):
# #            offset = 16*hdr["instr_pptr/offset[%u]" % index].value
# #            max_offset = max(offset+instr_size, max_offset)
# #            addr = self.absolute_address + 8*offset


class PTMModule(Module):
    PARSER_TAGS = {
        "id": "ptm",
        "category": "audio",
        "file_ext": ("ptm",),
        "min_size": 64 * 8,
        "description": "PolyTracker module (v1.17)"
    }
    MARKER = b"PTMF"
    HEADER = PTMHeader



class frame_view_imp_t:
    def on_frame_view_ready(self, dispatcher, frame_view):
        assert frame_view is not None
        self.view = frame_view

    def on_file_ready(self, dispatcher, file):
        assert file is not None
        self.filename = file.name

    def on_filename_update(self, dispatcher, filename):
        self.filename = filename

    def format_title(self, field):
        field_path = field.path
        return self.filename + '/' + field_path[1:]

    def on_field_activated(self, dispatcher, field):
        self.view.SetTitle(self.format_title(field))

    def on_activated(self):
        self.dispatcher.trigger('frame_activated', self.view)
